# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lightweight library for working with optimizers."""

from typing import Callable, List, Optional, Tuple, Union

from fedjax.core import dataclasses
from fedjax.core.typing import OptState
from fedjax.core.typing import Params

import haiku as hk
import jax
import optax

Grads = Params


@dataclasses.dataclass
class Optimizer:
  """Wraps different optimizer libraries in a common interface.

  Works with `optax <https://github.com/deepmind/optax>`_.

  The expected usage of Optimizer is as follows::

    # One step of SGD.
    params = {'w': jnp.array([1, 1, 1])}
    grads = {'w': jnp.array([2, 3, 4])}
    optimizer = fedjax.optimizers.sgd(learning_rate=0.1)
    opt_state = optimizer.init(params)
    opt_state, params = optimizer.apply(grads, opt_state, params)
    print(params)
    # {'w': DeviceArray([0.8, 0.7, 0.6], dtype=float32)}

  Attributes:
    init: Initializes (possibly empty) PyTree of statistics (optimizer state)
      given the input model parameters.
    apply: Transforms and applies the input gradients to update the optimizer
      state and model parameters.
  """
  init: Callable[[Params], OptState]
  apply: Callable[[Grads, OptState, Params], Tuple[OptState, Params]]


def create_optimizer_from_optax(opt: optax.GradientTransformation) -> Optimizer:
  """Creates optimizer from optax gradient transformation chain."""

  @jax.jit
  def apply(grads, opt_state, params):
    updates, opt_state = opt.update(grads, opt_state, params)
    params = optax.apply_updates(params, updates)
    return opt_state, params

  return Optimizer(opt.init, apply)


def ignore_grads_haiku(optimizer: Optimizer,
                       non_trainable_names: List[Tuple[str, str]]) -> Optimizer:
  """Modifies ``optimizer`` to ignore gradients for ``non_trainable_names``.

  Non-trainable parameters will have their values set to ``None`` when passed as
  input into the Optimizer to prevent any updates.

  NOTE: This will only work with models implemented in haiku.

  Args:
    optimizer: Base Optimizer.
    non_trainable_names: List of tuples of haiku module names and names of
      given entries in the module data bundle (e.g. parameter name). This list
      of names will be used to select the non-trainable parameters.

  Returns:
    Optimizer that will ignore gradients for the non-trainable parameters.
  """
  non_trainable_names = tuple(non_trainable_names)

  def non_trainable_to_none(module_name, name, value):
    if (module_name, name) in non_trainable_names:
      return None
    return value

  def init(params):
    trainable_params = hk.data_structures.map(non_trainable_to_none, params)
    return optimizer.init(trainable_params)

  def apply(grads, opt_state, params):
    trainable_grads = hk.data_structures.map(non_trainable_to_none, grads)
    trainable_params = hk.data_structures.map(non_trainable_to_none, params)
    opt_state, trainable_params = optimizer.apply(trainable_grads, opt_state,
                                                  trainable_params)
    # Set the non-trainable params from None back to their original values.
    trainable_params = hk.data_structures.to_mutable_dict(trainable_params)
    for module_name, name in non_trainable_names:
      trainable_params[module_name][name] = params[module_name][name]
    return opt_state, hk.data_structures.to_immutable_dict(trainable_params)

  return Optimizer(init, apply)


# Convenient aliases for `optax` optimizers.
# Docstrings copied directly from `optax`.
ScalarOrSchedule = Union[float, optax.Schedule]


def adagrad(learning_rate: ScalarOrSchedule,
            initial_accumulator_value: float = 0.1,
            eps: float = 1e-6) -> Optimizer:
  """The Adagrad optimizer.

  Adagrad is an algorithm for gradient based optimisation that anneals the
  learning rate for each parameter during the course of training.

  WARNING: Adagrad's main limit is the monotonic accumulation of squared
  gradients in the denominator: since all terms are >0, the sum keeps growing
  during training and the learning rate eventually becomes vanishingly small.

  References:
    [Duchi et al, 2011](https://jmlr.org/papers/v12/duchi11a.html)

  Args:
    learning_rate: This is a fixed global scaling factor.
    initial_accumulator_value: Initialisation for the accumulator.
    eps: A small constant applied to denominator inside of the square root (as
      in RMSProp) to avoid dividing by zero when rescaling.

  Returns:
    The corresponding `Optimizer`.
  """
  return create_optimizer_from_optax(
      optax.adagrad(
          learning_rate=learning_rate,
          initial_accumulator_value=initial_accumulator_value,
          eps=eps))


def adam(learning_rate: ScalarOrSchedule,
         b1: float = 0.9,
         b2: float = 0.999,
         eps: float = 1e-8,
         eps_root: float = 0.0) -> Optimizer:
  """The classic Adam optimiser.

  Adam is an SGD variant with learning rate adaptation. The `learning_rate`
  used for each weight is computed from estimates of first- and second-order
  moments of the gradients (using suitable exponential moving averages).

  References:
    [Kingma et al, 2014](https://arxiv.org/abs/1412.6980)

  Args:
    learning_rate: This is a fixed global scaling factor.
    b1: The exponential decay rate to track the first moment of past gradients.
    b2: The exponential decay rate to track the second moment of past gradients.
    eps: A small constant applied to denominator outside of the square root (as
      in the Adam paper) to avoid dividing by zero when rescaling.
    eps_root: A small constant applied to denominator inside the square root (as
      in RMSProp), to avoid dividing by zero when rescaling. This is needed for
      example when computing (meta-)gradients through Adam.

  Returns:
    The corresponding `Optimizer`.
  """
  return create_optimizer_from_optax(
      optax.adam(
          learning_rate=learning_rate, b1=b1, b2=b2, eps=eps,
          eps_root=eps_root))


def rmsprop(learning_rate: ScalarOrSchedule,
            decay: float = 0.9,
            eps: float = 1e-8,
            initial_scale: float = 0.,
            centered: bool = False,
            momentum: Optional[float] = None,
            nesterov: bool = False) -> Optimizer:
  """A flexible RMSProp optimiser.

  RMSProp is an SGD variant with learning rate adaptation. The `learning_rate`
  used for each weight is scaled by a suitable estimate of the magnitude of the
  gradients on previous steps. Several variants of RMSProp can be found
  in the literature. This alias provides an easy to configure RMSProp
  optimiser that can be used to switch between several of these variants.

  References:
    [Tieleman and Hinton, 2012](www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf)
    [Graves, 2013](https://arxiv.org/abs/1308.0850)

  Args:
    learning_rate: This is a fixed global scaling factor.
    decay: The decay used to track the magnitude of previous gradients.
    eps: A small numerical constant to avoid dividing by zero when rescaling.
    initial_scale: Initialisation of accumulators tracking the magnitude of
      previous updates. PyTorch uses `0`, TF1 uses `1`. When reproducing results
      from a paper, verify the value used by the authors.
    centered: Whether the second moment or the variance of the past gradients is
      used to rescale the latest gradients.
    momentum: The `decay` rate used by the momentum term, when it is set to
      `None`, then momentum is not used at all.
    nesterov: Whether nesterov momentum is used.

  Returns:
    The corresponding `Optimizer`.
  """
  return create_optimizer_from_optax(
      optax.rmsprop(
          learning_rate=learning_rate,
          decay=decay,
          eps=eps,
          initial_scale=initial_scale,
          centered=centered,
          momentum=momentum,
          nesterov=nesterov))


def sgd(learning_rate: ScalarOrSchedule,
        momentum: Optional[float] = None,
        nesterov: bool = False) -> Optimizer:
  """A canonical Stochastic Gradient Descent optimiser.

  This implements stochastic gradient descent. It also includes support for
  momentum, and nesterov acceleration, as these are standard practice when
  using stochastic gradient descent to train deep neural networks.

  References:
    [Sutskever et al, 2013](http://proceedings.mlr.press/v28/sutskever13.pdf)

  Args:
    learning_rate: This is a fixed global scaling factor.
    momentum: The `decay` rate used by the momentum term, when it is set to
      `None`, then momentum is not used at all.
    nesterov: Whether nesterov momentum is used.

  Returns:
    The corresponding `Optimizer`.
  """
  return create_optimizer_from_optax(
      optax.sgd(
          learning_rate=learning_rate, momentum=momentum, nesterov=nesterov))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.models.emnist."""

from fedjax.legacy.datasets import emnist as emnist_data
from fedjax.legacy.models import emnist as emnist_model
import haiku as hk
import numpy as np
import tensorflow as tf


def _mock_emnist_data(only_digits=True, seed=0):
  """Returns fixed example to avoid reading data over network in test."""
  num_classes = 10 if only_digits else 62
  dataset = tf.data.Dataset.from_tensor_slices({
      'pixels': np.random.RandomState(seed).random_sample((1, 28, 28)),
      'label': [np.random.RandomState(seed).randint(num_classes)],
  })
  return emnist_data.flip_and_expand(dataset)


class EmnistHaikuDenseTest(tf.test.TestCase):

  def create_model(self):
    return emnist_model.create_dense_model(only_digits=False)

  def setUp(self):
    super().setUp()
    self._rng_seq = hk.PRNGSequence(42)
    self._batch_size = 4
    self._batch = next(
        _mock_emnist_data(only_digits=False)
        .repeat(self._batch_size)
        .batch(self._batch_size)
        .as_numpy_iterator())
    self._model = self.create_model()

  def test_backward_pass(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    output = self._model.backward_pass(params, self._batch, next(self._rng_seq))
    self.assertGreaterEqual(output.loss, 0)
    self.assertEqual(output.num_examples, self._batch_size)

  def test_evaluate(self):
    params = self._model.init_params(rng=next(self._rng_seq))
    metrics = self._model.evaluate(params, self._batch)
    self.assertContainsSubset(['loss', 'accuracy', 'num_examples'],
                              metrics.keys())


class EmnistHaikuLogisticTest(EmnistHaikuDenseTest):

  def create_model(self):
    return emnist_model.create_logistic_model(only_digits=False)


class EmnistHaikuConvTest(EmnistHaikuDenseTest):

  def create_model(self):
    return emnist_model.create_conv_model(only_digits=False)


class EmnistStaxDenseTest(EmnistHaikuDenseTest):

  def create_model(self):
    return emnist_model.create_stax_dense_model(only_digits=False)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.aggregators.compression."""
from absl.testing import absltest

from fedjax.aggregators import compression
import jax
import jax.numpy as jnp
import numpy.testing as npt


class CompressionTest(absltest.TestCase):

  def test_num_leaves(self):
    params = {'w': jnp.array([1., 2., 3.]), 'b': jnp.array([1.])}
    self.assertEqual(compression.num_leaves(params), 2)

  def test_binary_stochastic_quantize_identity(self):
    # If the vector has only two distinct values, it should not change.
    v = jnp.array([0., 2., 2.])
    rng = jax.random.PRNGKey(42)
    compressed_v = compression.binary_stochastic_quantize(v, rng)
    npt.assert_array_equal(compressed_v, v)

  def test_binary_stochastic_quantize_unbiasedness(self):
    # If the vector has only two distinct values, it should not change.
    v = jnp.array([0., 1., 2.])
    rng = jax.random.PRNGKey(42)
    compressed_sum = jnp.zeros_like(v)
    for _ in range(500):
      rng, use_rng = jax.random.split(rng)
      compressed_sum += compression.binary_stochastic_quantize(v, use_rng)
    npt.assert_array_almost_equal(compressed_sum / 500, v, decimal=2)

  def test_uniform_stochastic_quantize_identity(self):
    # If the vector has only two distinct values, it should not change.
    v = jnp.array([0., 2., 2., 4.])
    rng = jax.random.PRNGKey(42)
    compressed_v = compression.uniform_stochastic_quantize(v, 3, rng)
    npt.assert_array_equal(compressed_v, v)

  def test_uniform_stochastic_quantize_all_equal(self):
    # If the vector has only two distinct values, it should not change.
    v = jnp.array([4., 4., 4., 4.])
    rng = jax.random.PRNGKey(42)
    compressed_v = compression.uniform_stochastic_quantize(v, 4, rng)
    npt.assert_array_equal(compressed_v, v)

  def test_uniform_stochastic_quantize_unbiasedness_one_dim(self):
    # If the vector has only three distinct values, it should not change.
    v = jnp.array([0., 1., 100.])
    rng = jax.random.PRNGKey(42)
    compressed_sum = jnp.zeros_like(v)
    for _ in range(500):
      rng, use_rng = jax.random.split(rng)
      compressed_sum += compression.uniform_stochastic_quantize(v, 125, use_rng)
    npt.assert_array_almost_equal(compressed_sum / 500, v, decimal=2)

  def test_uniform_stochastic_quantize_unbiasedness_two_dim(self):
    # If the vector has only three distinct values, it should not change.
    v = jnp.array([[0., 1., 100.], [0.3, 2.3, 45.]])
    rng = jax.random.PRNGKey(42)
    compressed_sum = jnp.zeros_like(v)
    for _ in range(500):
      rng, use_rng = jax.random.split(rng)
      compressed_sum += compression.uniform_stochastic_quantize(v, 125, use_rng)
    npt.assert_array_almost_equal(compressed_sum / 500, v, decimal=1)

  def test_uniform_stochastic_quantizer(self):
    delta_params_and_weights = [('a', {
        'w': jnp.array([1., 2., 3.])
    }, 2.), ('b', {
        'w': jnp.array([2., 4., 6.])
    }, 4.), ('c', {
        'w': jnp.array([1., 3., 5.])
    }, 2.)]

    quantizer = compression.uniform_stochastic_quantizer(
        3, jax.random.PRNGKey(0))
    init_aggregator_state = quantizer.init()
    quantized_params, new_state = quantizer.apply(delta_params_and_weights,
                                                  init_aggregator_state)
    self.assertEqual(new_state.num_bits, 67.0)
    npt.assert_array_equal(quantized_params['w'], [1.5, 3.25, 5.])


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for working with tree-like container data structures."""

from typing import Iterable, Iterator, List, Optional, Tuple, TypeVar

import jax
import jax.numpy as jnp

T = TypeVar('T')
tree_map = jax.tree_util.tree_map
tree_multimap = jax.tree_util.tree_multimap
# Dummy pmapped function to replicate an input pytree of DeviceArrays to
# ShardedDeviceArrays. The second argument can be any array as long as the
# leading axis has the correct size.
_p_broadcast = jax.pmap(lambda x, _: x, in_axes=(None, 0))


def tree_broadcast(pytree: T, axis_size: Optional[int] = None) -> T:
  axis_size = axis_size or jax.device_count()
  return _p_broadcast(pytree, jnp.arange(axis_size))


@jax.jit
def tree_stack(pytrees: List[T]) -> T:
  """Stacks input pytrees along leading axis into a single pytree."""
  return jax.tree_multimap(lambda *args: jnp.stack(args), *pytrees)


def tree_unstack(pytree: T, axis_size: Optional[int] = None) -> Iterator[T]:
  axis_size = axis_size or jax.device_count()
  for i in range(axis_size):
    yield jax.tree_map(lambda l, i_=i: l[i_], pytree)


@jax.jit
def tree_weight(pytree: T, weight: float) -> T:
  """Weights tree leaves by weight."""
  return jax.tree_map(lambda l: l * weight, pytree)


@jax.jit
def tree_zeros_like(pytree: T) -> T:
  """Creates a tree with zeros with same structure as the input."""
  return jax.tree_map(jnp.zeros_like, pytree)


@jax.jit
def tree_sum(*pytrees: T) -> T:
  """Sums input trees together."""
  sum_tree = pytrees[0]
  for tree in pytrees[1:]:
    sum_tree = jax.tree_multimap(lambda a, b: a + b, sum_tree, tree)
  return sum_tree


def tree_mean(pytrees_and_weights: Iterable[Tuple[T, float]]) -> T:
  """Returns (weighted) mean of input trees and weights.

  Args:
    pytrees_and_weights: Iterable of tuples of pytrees and associated weights.

  Returns:
    (Weighted) mean of input trees and weights.
  """
  sum_weighted_pytree = None
  sum_weight = 0.
  for pytree, weight in pytrees_and_weights:
    weighted_pytree = tree_weight(pytree, weight)
    if sum_weighted_pytree is None:
      sum_weighted_pytree = weighted_pytree
    else:
      sum_weighted_pytree = tree_sum(sum_weighted_pytree, weighted_pytree)
    sum_weight += weight
  inverse_weight = (1. / sum_weight) if sum_weight > 0. else 0.
  return tree_weight(sum_weighted_pytree, inverse_weight)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for structured_flags."""

from absl.testing import absltest
from absl.testing import flagsaver
from fedjax.core import client_datasets
from fedjax.core import optimizers
from fedjax.training import federated_experiment
from fedjax.training import structured_flags


class OptimizerFlagsTest(absltest.TestCase):

  OPTIMIZER = structured_flags.OptimizerFlags()
  NAMED_OPTIMIZER = structured_flags.OptimizerFlags('client')

  def test_get(self):
    for optimizer in structured_flags.OptimizerFlags.SUPPORTED:
      with self.subTest('get ' + optimizer):
        with flagsaver.flagsaver(optimizer=optimizer):
          self.assertIsInstance(self.OPTIMIZER.get(), optimizers.Optimizer)
        with flagsaver.flagsaver(client_optimizer=optimizer):
          self.assertIsInstance(self.NAMED_OPTIMIZER.get(),
                                optimizers.Optimizer)
    with self.subTest('get foo'):
      with self.assertRaisesRegex(ValueError, 'from --optimizer'):
        with flagsaver.flagsaver(optimizer='foo'):
          self.OPTIMIZER.get()
      with self.assertRaisesRegex(ValueError, 'from --client_optimizer'):
        with flagsaver.flagsaver(client_optimizer='foo'):
          self.NAMED_OPTIMIZER.get()


class AllBatchHParamsFlagsTest(absltest.TestCase):

  SHUFFLE_BATCH_REPEAT = structured_flags.ShuffleRepeatBatchHParamsFlags(
      'shuffle_batch_repeat')
  PADDED_BATCH = structured_flags.PaddedBatchHParamsFlags('padded')
  BATCH = structured_flags.BatchHParamsFlags()

  def test_get(self):
    with self.subTest('default'):
      self.assertEqual(
          self.SHUFFLE_BATCH_REPEAT.get(),
          client_datasets.ShuffleRepeatBatchHParams(batch_size=128))
      self.assertEqual(self.PADDED_BATCH.get(),
                       client_datasets.PaddedBatchHParams(batch_size=128))
      self.assertEqual(self.BATCH.get(),
                       client_datasets.BatchHParams(batch_size=128))
    with self.subTest('custom'):
      with flagsaver.flagsaver(
          shuffle_batch_repeat_batch_size=12,
          shuffle_batch_repeat_num_epochs=21,
          padded_batch_size=34,
          batch_size=56):
        self.assertEqual(
            self.SHUFFLE_BATCH_REPEAT.get(),
            client_datasets.ShuffleRepeatBatchHParams(
                batch_size=12, num_epochs=21))
        self.assertEqual(self.PADDED_BATCH.get(),
                         client_datasets.PaddedBatchHParams(batch_size=34))
        self.assertEqual(self.BATCH.get(),
                         client_datasets.BatchHParams(batch_size=56))


class FederatedExperimentConfigTest(absltest.TestCase):

  FEDERATED_EXPERIMENT_CONFIG = structured_flags.FederatedExperimentConfigFlags(
  )

  @flagsaver.flagsaver(root_dir='foo', num_rounds=1234)
  def test_get(self):
    with self.subTest('default'):
      self.assertEqual(
          self.FEDERATED_EXPERIMENT_CONFIG.get(),
          federated_experiment.FederatedExperimentConfig(
              root_dir='foo', num_rounds=1234))
    with self.subTest('custom'):
      with flagsaver.flagsaver(
          root_dir='bar',
          num_rounds=567,
          checkpoint_frequency=2,
          num_checkpoints_to_keep=3,
          eval_frequency=4):
        self.assertEqual(
            self.FEDERATED_EXPERIMENT_CONFIG.get(),
            federated_experiment.FederatedExperimentConfig(
                root_dir='bar',
                num_rounds=567,
                checkpoint_frequency=2,
                num_checkpoints_to_keep=3,
                eval_frequency=4))


class TaskFlagsTest(absltest.TestCase):

  TASK = structured_flags.TaskFlags()

  def test_get(self):
    with self.subTest('default'):
      with self.assertRaisesRegex(ValueError, 'Unsupported task: None'):
        self.TASK.get()


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.core.serialization."""

import os

from fedjax.legacy.core import serialization
from fedjax.legacy.core import test_util
import jax
import tensorflow as tf


class SerializationTest(tf.test.TestCase):

  def test_save_state(self):
    temp_dir = self.get_temp_dir()
    path = os.path.join(temp_dir, 'state')
    state = test_util.create_mock_state()

    serialization.save_state(state, path)

    self.assertTrue(tf.io.gfile.exists(path))

  def test_load_state(self):
    temp_dir = self.get_temp_dir()
    path = os.path.join(temp_dir, 'state')
    init_state = test_util.create_mock_state()
    serialization.save_state(init_state, path)

    state = serialization.load_state(path)

    expected_flat, expected_tree_def = jax.tree_flatten(init_state)
    actual_flat, actual_tree_def = jax.tree_flatten(state)
    for expected_array, actual_array in zip(expected_flat, actual_flat):
      self.assertAllEqual(expected_array, actual_array)
    self.assertEqual(expected_tree_def, actual_tree_def)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Toy regression models."""

import collections

from fedjax.legacy import core
import haiku as hk
import jax.numpy as jnp
import numpy as np


def create_regression_model() -> core.Model:
  """Creates toy regression model."""

  def forward_pass(batch):
    network = hk.Sequential([hk.Linear(1, with_bias=False)])
    return jnp.mean(network(batch['x']))

  def loss(batch, preds):
    return core.MeanMetric(
        total=jnp.square(jnp.mean(batch['y']) - preds), count=1.0)

  transformed_forward_pass = hk.transform(forward_pass)
  sample_batch = collections.OrderedDict(x=np.zeros((1, 1)), y=np.zeros((1,)))
  return core.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=sample_batch,
      loss_fn=loss)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EMNIST models."""

import collections
from typing import Callable, Optional

from fedjax.legacy import core
import haiku as hk
import jax
from jax.experimental import stax
import jax.numpy as jnp
import numpy as np


class Dropout(hk.Module):
  """Dropout haiku module."""

  def __init__(self, rate: float = 0.5):
    """Initializes dropout module.

    Args:
      rate: Probability that each element of x is discarded. Must be in [0, 1).
    """
    super().__init__()
    self._rate = rate

  def __call__(self, x: jnp.ndarray, is_train: bool):
    if is_train:
      return hk.dropout(rng=hk.next_rng_key(), rate=self._rate, x=x)
    return x


class ConvDropoutModule(hk.Module):
  """Custom haiku module for CNN with dropout.

  This must be defined as a custom hk.Module because only a single positional
  argument is allowed when using hk.Sequential.
  """

  def __init__(self, num_classes):
    super().__init__()
    self._num_classes = num_classes

  def __call__(self, x: jnp.ndarray, is_train: bool):
    x = hk.Conv2D(output_channels=32, kernel_shape=(3, 3), padding='VALID')(x)
    x = jax.nn.relu(x)
    x = hk.Conv2D(output_channels=64, kernel_shape=(3, 3), padding='VALID')(x)
    x = jax.nn.relu(x)
    x = (
        hk.MaxPool(
            window_shape=(1, 2, 2, 1), strides=(1, 2, 2, 1),
            padding='VALID')(x))
    x = Dropout(rate=0.25)(x, is_train)
    x = hk.Flatten()(x)
    x = hk.Linear(128)(x)
    x = jax.nn.relu(x)
    x = Dropout(rate=0.5)(x, is_train)
    x = hk.Linear(self._num_classes)(x)
    return x


# Defines the expected structure of input batches to the model. This is used to
# determine the model parameter shapes.
_EMNIST_HAIKU_SAMPLE_BATCH = collections.OrderedDict(
    x=np.zeros((1, 28, 28, 1)), y=np.zeros(1,))
_EMNIST_STAX_SAMPLE_SHAPE = (-1, 28, 28, 1)


def _loss(batch: core.Batch, preds: jnp.ndarray) -> core.Metric:
  return core.metrics.cross_entropy_loss_fn(targets=batch['y'], preds=preds)


def _accuracy(batch: core.Batch, preds: jnp.ndarray) -> core.Metric:
  return core.metrics.accuracy_fn(targets=batch['y'], preds=preds)


# Common definitions for EMNIST image recognition task.
_EMNIST_LOSS_FN = _loss
_EMNIST_METRICS_FN_MAP = collections.OrderedDict(accuracy=_accuracy)


def create_conv_model(
    only_digits: bool = False,
    reg_fn: Optional[Callable[[core.Params],
                              jnp.ndarray]] = None) -> core.Model:
  """Creates EMNIST CNN model with dropout."""
  num_classes = 10 if only_digits else 62

  def forward_pass(batch, is_train=True):
    return ConvDropoutModule(num_classes)(batch['x'], is_train)

  transformed_forward_pass = hk.transform(forward_pass)
  return core.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_EMNIST_HAIKU_SAMPLE_BATCH,
      loss_fn=_EMNIST_LOSS_FN,
      reg_fn=reg_fn,
      metrics_fn_map=_EMNIST_METRICS_FN_MAP,
      train_kwargs={'is_train': True},
      test_kwargs={'is_train': False})


def create_dense_model(
    only_digits: bool = False,
    hidden_units: int = 200,
    reg_fn: Optional[Callable[[core.Params],
                              jnp.ndarray]] = None) -> core.Model:
  """Creates EMNIST dense net."""
  num_classes = 10 if only_digits else 62

  def forward_pass(batch):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(hidden_units),
        jax.nn.relu,
        hk.Linear(hidden_units),
        jax.nn.relu,
        hk.Linear(num_classes),
    ])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  return core.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_EMNIST_HAIKU_SAMPLE_BATCH,
      loss_fn=_EMNIST_LOSS_FN,
      reg_fn=reg_fn,
      metrics_fn_map=_EMNIST_METRICS_FN_MAP)


def create_logistic_model(
    only_digits: bool = False,
    reg_fn: Optional[Callable[[core.Params],
                              jnp.ndarray]] = None) -> core.Model:
  """Creates EMNIST logistic model."""
  num_classes = 10 if only_digits else 62

  def forward_pass(batch):
    network = hk.Sequential([
        hk.Flatten(),
        hk.Linear(num_classes),
    ])
    return network(batch['x'])

  transformed_forward_pass = hk.transform(forward_pass)
  return core.create_model_from_haiku(
      transformed_forward_pass=transformed_forward_pass,
      sample_batch=_EMNIST_HAIKU_SAMPLE_BATCH,
      loss_fn=_EMNIST_LOSS_FN,
      reg_fn=reg_fn,
      metrics_fn_map=_EMNIST_METRICS_FN_MAP)


def create_stax_dense_model(
    only_digits: bool = False,
    hidden_units: int = 200,
    reg_fn: Optional[Callable[[core.Params],
                              jnp.ndarray]] = None) -> core.Model:
  """Creates EMNIST dense net via stax."""
  num_classes = 10 if only_digits else 62
  stax_init_fn, stax_apply_fn = stax.serial(stax.Flatten,
                                            stax.Dense(hidden_units), stax.Relu,
                                            stax.Dense(hidden_units), stax.Relu,
                                            stax.Dense(num_classes))
  return core.create_model_from_stax(
      stax_init_fn=stax_init_fn,
      stax_apply_fn=stax_apply_fn,
      sample_shape=_EMNIST_STAX_SAMPLE_SHAPE,
      loss_fn=_EMNIST_LOSS_FN,
      reg_fn=reg_fn,
      metrics_fn_map=_EMNIST_METRICS_FN_MAP)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Methods for serializing and deserializing arbitrary state pytrees."""

import pickle
from typing import Any

from absl import logging

import tensorflow as tf


def save_state(state: Any, path: str):
  """Saves state to file path."""
  logging.info('Saving state to %s.', path)
  with tf.io.gfile.GFile(path, 'wb') as f:
    pickle.dump(state, f)


def load_state(path: str) -> Any:
  """Loads saved state from file path."""
  logging.info('Loading params from %s.', path)
  with tf.io.gfile.GFile(path, 'rb') as f:
    return pickle.load(f)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lightweight convenience container for various model implementations."""

import functools
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Mapping, Tuple

from fedjax.core import client_datasets
from fedjax.core import dataclasses
from fedjax.core import federated_data
from fedjax.core import for_each_client
from fedjax.core import metrics
from fedjax.core import util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import BatchPrediction
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey

import haiku as hk
import jax
import jax.numpy as jnp

# Typically unnormalized model forward pass output.
BatchTrainOutput = jnp.ndarray
BatchEvalPrediction = BatchPrediction


@dataclasses.dataclass
class Model:
  """Container class for models.

  Model exists to provide easy access to predefined neural network models.
  It is meant to contain all the information needed for standard centralized
  training and evaluation. Non-standard training methods can be built upon the
  information avaiable in Model along with any additional information
  (e.g. interpolation can be implemented as a composition of two models along
  with an interpolation weight).

  Works for Haiku and jax.experimental.stax.

  The expected usage of Model is as follows::

    # Training.
    step_size = 0.1
    rng = jax.random.PRNGKey(0)
    params = model.init(rng)

    def loss(params, batch, rng):
      preds = model.apply_for_train(params, batch, rng)
      return jnp.sum(model.train_loss(batch, preds))

    grad_fn = jax.grad(loss)
    for batch in batches:
      rng, use_rng = jax.random.split(rng)
      grads = grad_fn(params, batch, use_rng)
      params = jax.tree_util.tree_multimap(lambda a, b: a - step_size * b,
                                           params, grads)

    # Evaluation.
    print(fedjax.evaluate_model(model, params, batches))
    # Example output:
    # {'loss': 2.3, 'accuracy': 0.2}

  The following is an example using Model compositionally as a building block
  to impelement model interpolation::

    def interpolate(model_1, model_2, init_weight):

      @jax.jit
      def init(rng):
        rng_1, rng_2 = jax.random.split(rng)
        params_1 = model_1.init(rng_1)
        params_2 = model_2.init(rng_2)
        return params_1, params_2, init_weight

      @jax.jit
      def apply_for_train(params, input, rng):
        rng_1, rng_2 = jax.random.split(rng)
        params_1, params_2, weight = params
        return (model_1.apply_for_train(params_1, input, rng_1) * weight +
                model_2.apply_for_train(params_1, input, rng_2) * (1 - weight))

      @jax.jit
      def apply_for_eval(params, input):
        params_1, params_2, weight = params
        return (model_1.apply_for_eval(params_1, input) * weight +
                model_2.apply_for_eval(params_2, input) * (1 - weight))

      return fedjax.Model(init,
                          apply_for_train,
                          apply_for_eval,
                          model_1.train_loss,
                          model_1.eval_metrics)

    model = interpolate(model_1, model_2, init_weight=0.5)

  Attributes:
    init: Initialization function that takes a seed PRNGKey and returns
      a PyTree of initialized parameters (i.e. model weights). These parameters
      will be passed as input into :meth:`apply_for_train` and
      :meth:`apply_for_eval`.
      Any trainable weights for a model that are modified in the training loop
      should be contained inside of these parameters.
    apply_for_train: Function that takes the parameters PyTree, batch of
      examples, and PRNGKey as inputs and outputs the model predictions for
      training that are then passed into :meth:`train_loss`.
      This considers strategies such as dropout.
    apply_for_eval: Function that usually takes the parameters PyTree and batch
      of examples as inputs and outputs the model predictions for evaluation
      that are then passed to :attr:`eval_metrics`.
      This is defined separately from :meth:`apply_for_train` to avoid
      having to specify inputs like PRNGKey that are not used in evaluation.
    train_loss: Loss function for training that takes batch of examples and
      model output from :meth:`apply_for_train` as input that outputs per
      example loss. This will typically called inside a :func:`jax.grad` wrapped
      function to compute gradients.
    eval_metrics: Ordered mapping of evaluation metric names to
      :class:`~fedjax.metrics.Metric`. These :class:`~fedjax.metrics.Metric` s
      are defined for single examples and will be used in :func:`evaluate_model`
  """
  init: Callable[[PRNGKey], Params]
  apply_for_train: Callable[[Params, BatchExample, PRNGKey], BatchTrainOutput]
  apply_for_eval: Callable[[Params, BatchExample], BatchEvalPrediction]
  train_loss: Callable[[BatchExample, BatchTrainOutput], jnp.ndarray]
  eval_metrics: Mapping[str, metrics.Metric]

  # Prevent dataclass from creating hash/eq so that a Model object remains
  # id hashed. This allows eval_metrics to be a standard dict.
  def __hash__(self) -> int:
    return id(self)

  def __eq__(self, other: Any) -> bool:
    return self is other


def create_model_from_haiku(
    transformed_forward_pass: hk.Transformed,
    sample_batch: BatchExample,
    train_loss: Callable[[BatchExample, BatchTrainOutput], jnp.ndarray],
    eval_metrics: Optional[Mapping[str, metrics.Metric]] = None,
    train_kwargs: Optional[Mapping[str, Any]] = None,
    eval_kwargs: Optional[Mapping[str, Any]] = None) -> Model:
  """Creates Model after applying defaults and haiku specific preprocessing.

  Args:
    transformed_forward_pass: Transformed forward pass from :func:`hk.transform`
    sample_batch: Example input used to determine model parameter shapes.
    train_loss: Loss function for training that outputs per example loss.
    eval_metrics: Mapping of evaluation metric names to
      :class:`~fedjax.metrics.Metric`. These metrics are defined for
      single examples and will be consumed in :func:`evaluate_model`.
    train_kwargs: Keyword arguments passed to model for training.
    eval_kwargs: Keyword arguments passed to model for evaluation.

  Returns:
    Model
  """
  eval_metrics = eval_metrics or {}
  train_kwargs = train_kwargs or {}
  eval_kwargs = eval_kwargs or {}

  @jax.jit
  def init(rng):
    return transformed_forward_pass.init(rng, sample_batch)

  @jax.jit
  def apply_for_train(params, batch, rng=None):
    return transformed_forward_pass.apply(params, rng, batch, **train_kwargs)

  @jax.jit
  def apply_for_eval(params, batch):
    return transformed_forward_pass.apply(params, None, batch, **eval_kwargs)

  return Model(init, apply_for_train, apply_for_eval, train_loss, eval_metrics)


def create_model_from_stax(
    stax_init: Callable[..., Params],
    stax_apply: Callable[..., jnp.ndarray],
    sample_shape: Tuple[int, ...],
    train_loss: Callable[[BatchExample, BatchTrainOutput], jnp.ndarray],
    eval_metrics: Optional[Mapping[str, metrics.Metric]] = None,
    train_kwargs: Optional[Mapping[str, Any]] = None,
    eval_kwargs: Optional[Mapping[str, Any]] = None,
    input_key: str = 'x') -> Model:
  """Creates Model after applying defaults and stax specific preprocessing.

  Args:
    stax_init: Initialization function returned from :func:`stax.serial`.
    stax_apply: Model forward_pass pass function returned from stax.serial.
    sample_shape: The expected shape of the input to the model.
    train_loss: Loss function for training that outputs per example loss.
    eval_metrics: Mapping of evaluation metric names to
      :class:`~fedjax.metrics.Metric`. These metrics are defined for
      single examples and will be consumed in :func:`evaluate_model`.
    train_kwargs: Keyword arguments passed to model for training.
    eval_kwargs: Keyword arguments passed to model for evaluation.
    input_key: Key name for the input in batch mapping.

  Returns:
    Model
  """
  eval_metrics = eval_metrics or {}
  train_kwargs = train_kwargs or {}
  eval_kwargs = eval_kwargs or {}

  @jax.jit
  def init(rng):
    _, params = stax_init(rng, sample_shape)
    return params

  @jax.jit
  def apply_for_train(params, batch, rng=None):
    return stax_apply(params, batch[input_key], rng=rng, **train_kwargs)

  @jax.jit
  def apply_for_eval(params, batch):
    return stax_apply(params, batch[input_key], **eval_kwargs)

  return Model(init, apply_for_train, apply_for_eval, train_loss, eval_metrics)


@functools.partial(jax.jit, static_argnums=0)
def _evaluate_model_step(model: Model, params: Params, batch: BatchExample,
                         stat: metrics.Stat) -> Dict[str, metrics.Stat]:
  """Evaluates model for one batch and returns merged Stat.

  Args:
    model: Model container with apply_for_eval and eval_metrics.
    params: Pytree of model parameters to be evaluated.
    batch: Batch of N examples.
    stat: Intermediate Stat from the previous step to be accumulated in the
      current step.

  Returns:
    A dictionary of intermediate evaluation Stats.
  """
  try:
    mask = batch[client_datasets.EXAMPLE_MASK_KEY].astype(jnp.bool_)
  except KeyError:
    mask = jnp.ones([len(next(iter(batch.values())))], dtype=jnp.bool_)
  pred = model.apply_for_eval(params, batch)
  new_stat = {
      k: metrics.evaluate_batch(metric, batch, pred, mask)
      for k, metric in model.eval_metrics.items()
  }
  return jax.tree_util.tree_multimap(
      lambda a, b: a.merge(b),
      stat,
      new_stat,
      is_leaf=lambda v: isinstance(v, metrics.Stat))


def evaluate_model(model: Model, params: Params,
                   batches: Iterable[BatchExample]) -> Dict[str, jnp.ndarray]:
  """Evaluates model for multiple batches and returns final results.

  This is the recommended way to compute evaluation metrics for a given model.

  Args:
    model: Model container.
    params: Pytree of model parameters to be evaluated.
    batches: Multiple batches to compute and aggregate evaluation metrics over.
      Each batch can optional contain a feature keyed by
      client_datasets.MASK_KEY (see :meth:`ClientDataset.padded_batch` ).

  Returns:
    A dictionary of evaluation :class:`~fedjax.metrics.Metric` results.
  """
  stat = {k: metric.zero() for k, metric in model.eval_metrics.items()}
  for batch in batches:
    stat = _evaluate_model_step(model, params, batch, stat)
  return jax.tree_util.tree_map(
      lambda x: x.result(), stat, is_leaf=lambda v: isinstance(v, metrics.Stat))


class ModelEvaluator:
  """Evaluates model for each client dataset, either using global params, or per client params.

  To evaluate a Model on a single dataset, use evaluate_model() instead.
  """

  def __init__(self, model: Model):
    # params can be passed in 2 ways:
    # -   As `shared_input`: All clients are evaluated using the same params.
    # -   As `client_input`: Each client is evaluated using per client params.
    def client_init(shared_input, client_input):
      if shared_input is not None:
        params = shared_input
      else:
        params = client_input
      stat = {k: metric.zero() for k, metric in model.eval_metrics.items()}
      return params, stat

    def client_step(state, batch):
      params, stat = state
      next_stat = _evaluate_model_step(model, params, batch, stat)
      return params, next_stat

    def client_final(shared_input, state):
      del shared_input
      _, stat = state
      return {k: v.result() for k, v in stat.items()}

    self._evaluate_each_client = for_each_client.for_each_client(
        client_init, client_step, client_final)

  def evaluate_global_params(
      self, params: Params, clients: Iterable[Tuple[federated_data.ClientId,
                                                    Iterable[BatchExample]]]
  ) -> Iterator[Tuple[federated_data.ClientId, Dict[str, jnp.ndarray]]]:
    """Evaluates batches from each client using global params.

    Args:
      params: Model params to evaluate.
      clients: Client batches.

    Yields:
      Pairs of the client id and a dictionary of evaluation `Metric` results for
      each client.
    """
    yield from self._evaluate_each_client(
        shared_input=params,
        clients=[(client_id, batches, None) for client_id, batches in clients])

  def evaluate_per_client_params(
      self, clients: Iterable[Tuple[federated_data.ClientId,
                                    Iterable[BatchExample], Params]]
  ) -> Iterator[Tuple[federated_data.ClientId, Dict[str, jnp.ndarray]]]:
    """Evaluates batches from each client using per client params.

    Args:
      clients: Client batches and the per client params.

    Yields:
      Pairs of the client id and a dictionary of evaluation `Metric` results for
      each client.
    """
    yield from self._evaluate_each_client(shared_input=None, clients=clients)


def model_per_example_loss(
    model: Model) -> Callable[[Params, BatchExample, PRNGKey], jnp.ndarray]:
  """Convenience function for constructing a per-example loss function from a model.

  Args:
    model: Model.

  Returns:
    A function from (params, batch_example, rng) to a vector of loss values for
    each example in the batch.
  """

  def per_example_loss(params, batch_example, rng):
    train_output = model.apply_for_train(params, batch_example, rng)
    return model.train_loss(batch_example, train_output)

  return per_example_loss


@functools.partial(jax.jit, static_argnums=0)
def _evaluate_average_loss_step(per_example_loss, params, batch, rng,
                                accum_loss, num_examples):
  """Evaluates average per example loss for one batch and returns updated accumlators."""
  rng, use_rng = jax.random.split(rng)
  loss = per_example_loss(params, batch, use_rng)
  if client_datasets.EXAMPLE_MASK_KEY in batch:
    mask = batch[client_datasets.EXAMPLE_MASK_KEY]
    accum_loss += jnp.vdot(mask, loss)
    num_examples += jnp.sum(mask)
  else:
    accum_loss += jnp.sum(loss)
    num_examples += len(loss)
  return rng, accum_loss, num_examples


@functools.partial(jax.jit, static_argnums=0)
def _finalize_average_loss(regularizer, params, accum_loss, num_examples):
  average_loss = util.safe_div(accum_loss, num_examples)
  if regularizer is not None:
    average_loss += regularizer(params)
  return average_loss


def evaluate_average_loss(
    params: Params,
    batches: Iterable[BatchExample],
    rng: PRNGKey,
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    regularizer: Optional[Callable[[Params],
                                   jnp.ndarray]] = None) -> jnp.ndarray:
  """Evaluates the average per example loss over multiple batches.

  Args:
    params: PyTree of model parameters to be evaluated.
    batches: Multiple batches to compute and aggregate evaluation metrics over.
      Each batch can optional contain a feature keyed by
      client_datasets.MASK_KEY (see ClientDataset.padded_batch).
    rng: Initial PRNGKey for making per_example_loss calls.
    per_example_loss: Per example loss function.
    regularizer: Optional regularizer function.

  Returns:
    The average per example loss, plus the regularizer term when specified.
  """
  accum_loss, num_examples = 0, 0
  for batch in batches:
    rng, accum_loss, num_examples = _evaluate_average_loss_step(
        per_example_loss=per_example_loss,
        params=params,
        batch=batch,
        rng=rng,
        accum_loss=accum_loss,
        num_examples=num_examples)
  return _finalize_average_loss(
      regularizer=regularizer,
      params=params,
      accum_loss=accum_loss,
      num_examples=num_examples)


class AverageLossEvaluator:
  """Evaluates average loss for each client dataset, either using global params, or per client params.

  The average loss is defined as the average per example loss, plus the
  regularizer term when specified. To evaluate average loss on a single dataset,
  use evaluate_average_loss() instead.
  """

  def __init__(self,
               per_example_loss: Callable[[Params, BatchExample, PRNGKey],
                                          jnp.ndarray],
               regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
    # params can be passed in 2 ways:
    # -   As `shared_input`: All clients are evaluated using the same params.
    # -   As `client_input`: Each client is evaluated using per client params.
    def client_init(shared_input, client_input):
      if shared_input is not None:
        params = shared_input
        rng = client_input
      else:
        rng, params = client_input
      accum_loss = 0.
      num_examples = 0.
      return rng, params, accum_loss, num_examples

    def client_step(state, batch):
      rng, params, accum_loss, num_examples = state
      rng, accum_loss, num_examples = _evaluate_average_loss_step(
          per_example_loss=per_example_loss,
          params=params,
          batch=batch,
          rng=rng,
          accum_loss=accum_loss,
          num_examples=num_examples)
      return rng, params, accum_loss, num_examples

    def client_final(shared_input, state):
      del shared_input
      _, params, accum_loss, num_examples = state
      return _finalize_average_loss(
          regularizer=regularizer,
          params=params,
          accum_loss=accum_loss,
          num_examples=num_examples)

    self._evaluate_each_client = for_each_client.for_each_client(
        client_init, client_step, client_final)

  def evaluate_global_params(
      self, params: Params, clients: Iterable[Tuple[federated_data.ClientId,
                                                    Iterable[BatchExample],
                                                    PRNGKey]]
  ) -> Iterator[Tuple[federated_data.ClientId, jnp.ndarray]]:
    """Evaluates batches from each client using global params.

    Args:
      params: Model params to evaluate.
      clients: Client batches.

    Yields:
      Pairs of the client id and the client's average loss.
    """
    yield from self._evaluate_each_client(shared_input=params, clients=clients)

  def evaluate_per_client_params(
      self, clients: Iterable[Tuple[federated_data.ClientId,
                                    Iterable[BatchExample], PRNGKey, Params]]
  ) -> Iterator[Tuple[federated_data.ClientId, jnp.ndarray]]:
    """Evaluates batches from each client using per client params.

    Args:
      clients: Client batches and the per client params.

    Yields:
      Pairs of the client id and the client's average loss.
    """
    yield from self._evaluate_each_client(
        shared_input=None,
        clients=[(client_id, batches, (rng, params))
                 for client_id, batches, rng, params in clients])


def grad(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None
) -> Callable[[Params, BatchExample, PRNGKey], Params]:
  """A standard gradient function derived from per-example loss and an optional regularizer.

  The scalar loss function being differentiated is simply:

    mean(per-example loss) + regularizer term

  The returned gradient function support both unpadded batches, and padded
  batches with the mask feature keyed by client_datasets.EXAMPLE_MASK_KEY.

  Args:
    per_example_loss: A function from (params, batch_example, rng) to a vector
      of loss values for each example in the batch.
    regularizer: Optional regularizer that only depends on params.

  Returns:
    A function from (params, batch_example, rng) to gradients.
  """

  def scalar_loss(params, batch_example, rng):
    batch_loss = per_example_loss(params, batch_example, rng)
    if client_datasets.EXAMPLE_MASK_KEY in batch_example:
      mask = batch_example[client_datasets.EXAMPLE_MASK_KEY]
      num_examples = jnp.sum(mask)
      loss = util.safe_div(jnp.vdot(batch_loss, mask), num_examples)
    else:
      loss = jnp.mean(batch_loss)
    if regularizer is not None:
      loss += regularizer(params)
    return loss

  return jax.jit(jax.grad(scalar_loss))


def model_grad(
    model: Model,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None
) -> Callable[[Params, BatchExample, PRNGKey], Params]:
  """A standard gradient function derived from a model and an optional regularizer.

  The scalar loss function being differentiated is simply:

    mean(model's per-example loss) + regularizer term

  The returned gradient function support both unpadded batches, and padded
  batches with the mask feature keyed by client_datasets.EXAMPLE_MASK_KEY.

  Args:
    model: A Model.
    regularizer: Optional regularizer.

  Returns:
    A function from (params, batch_example, rng) to gradients.
  """

  return grad(model_per_example_loss(model), regularizer)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated experiment manager."""

import os.path
import time
from typing import Any, Callable, List, Mapping, NamedTuple, Optional, TypeVar

from absl import logging
from fedjax.legacy import core
from fedjax.legacy.training import checkpoint
from fedjax.legacy.training import logging as fedjax_logging
import jax.numpy as jnp
import numpy as np
import tensorflow as tf


def set_tf_cpu_only():
  """Restricts TensorFlow device visibility to only CPU.

  TensorFlow is only used for data loading, so we prevent it from allocating
  GPU/TPU memory.
  """
  tf.config.experimental.set_visible_devices([], 'GPU')
  tf.config.experimental.set_visible_devices([], 'TPU')


def get_pseudo_random_state(round_num: int,
                            random_seed: Optional[int] = None
                           ) -> np.random.RandomState:
  """Computes a numpy random state as a function of round_num and random_seed."""
  #  Settings for a multiplicative linear congruential generator (aka Lehmer
  #  generator) suggested in 'Random Number Generators: Good
  #  Ones are Hard to Find' by Park and Miller.
  if isinstance(random_seed, int):
    mlcg_modulus = 2**(31) - 1
    mlcg_multiplier = 16807
    mlcg_start = np.random.RandomState(random_seed).randint(1, mlcg_modulus - 1)
    return np.random.RandomState(
        pow(mlcg_multiplier, round_num, mlcg_modulus) * mlcg_start %
        mlcg_modulus)
  return np.random.RandomState()


class FederatedExperimentConfig(NamedTuple):
  root_dir: str
  num_rounds: int
  num_clients_per_round: int
  sample_client_random_seed: Optional[int] = None
  checkpoint_frequency: int = 0
  num_checkpoints_to_keep: int = 1
  eval_frequency: int = 0


class ClientEvaluationFn:
  """Evaluation function for a subset of client(s)."""

  def __init__(self, federated_data: core.FederatedData, model: core.Model,
               config: FederatedExperimentConfig):
    self._federated_data = federated_data
    self._model = model
    self._num_clients_per_round = config.num_clients_per_round
    self._sample_client_random_seed = config.sample_client_random_seed

  def _sample_clients(self, round_num: int) -> List[str]:
    random_state = get_pseudo_random_state(round_num,
                                           self._sample_client_random_seed)
    return list(
        random_state.choice(
            self._federated_data.client_ids,
            size=self._num_clients_per_round,
            replace=False))

  def __call__(self, state: Any, round_num: int) -> core.MetricResults:
    client_ids = self._sample_clients(round_num)
    combined_dataset = core.create_tf_dataset_for_clients(
        self._federated_data, client_ids=client_ids)
    return core.evaluate_single_client(combined_dataset, self._model,
                                       state.params)


class FullEvaluationFn:
  """Evaluation function for all of client(s)."""

  def __init__(self, federated_data: core.FederatedData, model: core.Model):
    self._dataset = core.dataset_util.create_tf_dataset_for_clients(
        federated_data)
    self._model = model

  def __call__(self, state: Any, round_num: int) -> core.MetricResults:
    del round_num
    return core.evaluate_single_client(self._dataset, self._model, state.params)


def _block_until_ready_state(state: Any):
  """Recursively calls block_until_ready on valid pytree leaves of state."""

  def _safe_block_until_ready(l):
    if isinstance(l, jnp.ndarray):
      l.block_until_ready()

  core.tree_map(_safe_block_until_ready, state)


T = TypeVar('T')


# TODO(theertha): Add functionality for warmstart / custom initialization.
def run_federated_experiment(
    config: FederatedExperimentConfig,
    federated_algorithm: core.FederatedAlgorithm[T],
    periodic_eval_fn_map: Optional[Mapping[str, Callable[
        [T, int], core.MetricResults]]] = None,
    final_eval_fn_map: Optional[Mapping[str,
                                        Callable[[T, int],
                                                 core.MetricResults]]] = None
) -> T:
  """Runs federated algorithm experiment and auxiliary processes.

  Args:
    config: FederatedExperimentConfig configurations.
    federated_algorithm: FederatedAlgorithm to be run over multiple rounds.
    periodic_eval_fn_map: Mapping of name to evaluation functions that are run
      repeatedly over multiple federated training rounds. The frequency is
      defined in `_FederatedExperimentConfig.eval_frequency`.
    final_eval_fn_map: Mapping of name to evaluation functions that are run at
      the very end of federated training. Typically, full test evaluation
      functions will be set here.

  Returns:
    Final state of the input federated algortihm after training.
  """
  if config.root_dir:
    tf.io.gfile.makedirs(config.root_dir)

  if periodic_eval_fn_map is None:
    periodic_eval_fn_map = {}
  if final_eval_fn_map is None:
    final_eval_fn_map = {}

  logger = fedjax_logging.Logger(config.root_dir)

  latest = checkpoint.load_latest_checkpoint(config.root_dir)
  if latest:
    state, start_round_num = latest
  else:
    state = federated_algorithm.init_state()
    start_round_num = 1

  start = time.time()
  for round_num in range(start_round_num, config.num_rounds + 1):
    # Get a random state and randomly sample clients.
    random_state = get_pseudo_random_state(round_num,
                                           config.sample_client_random_seed)
    client_ids = list(
        random_state.choice(
            federated_algorithm.federated_data.client_ids,
            size=config.num_clients_per_round,
            replace=False))
    logging.info('round_num %d: client_ids = %s', round_num, client_ids)

    # Run one round of the algorithm, where bulk of the work happens.
    state = federated_algorithm.run_round(state, client_ids)

    # Save checkpoint.
    should_save_checkpoint = config.checkpoint_frequency and (
        round_num == start_round_num or
        round_num % config.checkpoint_frequency == 0)
    if should_save_checkpoint:
      checkpoint.save_checkpoint(config.root_dir, state, round_num,
                                 config.num_checkpoints_to_keep)

    # Run evaluation.
    should_run_eval = config.eval_frequency and (
        round_num == start_round_num or round_num % config.eval_frequency == 0)
    if should_run_eval:
      start_periodic_eval = time.time()
      for eval_name, eval_fn in periodic_eval_fn_map.items():
        metrics = eval_fn(state, round_num)
        if metrics:
          for metric_name, metric_value in metrics.items():
            logger.log(eval_name, metric_name, metric_value, round_num)
      logger.log('.', 'periodic_eval_duration_sec',
                 time.time() - start_periodic_eval, round_num)

    # Log the time it takes per round. Rough approximation since we're not
    # using DeviceArray.block_until_ready()
    logger.log('.', 'mean_round_duration_sec',
               (time.time() - start) / (round_num + 1 - start_round_num),
               round_num)

  # Logging overall time it took.
  # DeviceArray.block_until_ready() is needed for accurate timing due to
  # https://jax.readthedocs.io/en/latest/async_dispatch.html.
  _block_until_ready_state(state)
  num_rounds = config.num_rounds - start_round_num
  mean_round_duration = ((time.time() - start) /
                         num_rounds if num_rounds > 0 else 0)

  # Final evaluation.
  final_eval_start = time.time()
  for eval_name, eval_fn in final_eval_fn_map.items():
    metrics = eval_fn(state, round_num)
    if metrics:
      metrics_path = os.path.join(config.root_dir, f'{eval_name}.tsv')
      with tf.io.gfile.GFile(metrics_path, 'w') as f:
        f.write('\t'.join(metrics.keys()) + '\n')
        f.write('\t'.join([str(v) for v in metrics.values()]))
  # DeviceArray.block_until_ready() isn't needed here since we write to file.
  final_eval_duration = time.time() - final_eval_start
  logging.info('mean_round_duration = %f sec.', mean_round_duration)
  logging.info('final_eval_duration = %f sec.', final_eval_duration)
  return state



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.core.regularizers."""

from fedjax.legacy.core import regularizers
from fedjax.legacy.core import test_util
import jax
import tensorflow as tf


class RegularizerTest(tf.test.TestCase):

  def test_l2_regularizer(self):
    params = test_util.create_mock_state(seed=0).params
    output = regularizers.L2Regularizer()(params)
    self.assertAlmostEqual(output, 37.64189)

  def test_l2_regularizer_weight(self):
    params = test_util.create_mock_state(seed=0).params
    original_output = regularizers.L2Regularizer()(params)
    output = regularizers.L2Regularizer(weight=0.2)(params)
    self.assertAlmostEqual(output, original_output * 0.2)

  def test_l2_regularizer_parameter_weight(self):
    params = test_util.create_mock_state(seed=0).params
    original_output = regularizers.L2Regularizer()(params)

    param_weights = jax.tree_map(lambda leaf: 2*jax.numpy.ones(leaf.shape),
                                 params)
    output = regularizers.L2Regularizer(weight=1.0,
                                        param_weights=param_weights)(params)
    self.assertAlmostEqual(output, 2*original_output,
                           delta=1e-5)

  def test_l2_regularizer_evaluation_with_center(self):
    params = test_util.create_mock_state(seed=0).params
    original_output = regularizers.L2Regularizer()(params)
    center_params = jax.tree_map(lambda l: l * 0.2, params)
    output = regularizers.L2Regularizer(center_params=center_params)(params)
    self.assertAlmostEqual(output, original_output * (1. - 0.2)**2)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FedJAX types."""

from typing import Any, Mapping, Union
import jax.numpy as jnp
import optax

# The difference between Batch* and Single* is Single* is a mapping to a SINGLE
# example; where as Batch* is to a batch of N examples.
BatchExample = Mapping[str, jnp.ndarray]
SingleExample = Mapping[str, jnp.ndarray]

BatchPrediction = Union[jnp.ndarray, Mapping[str, jnp.ndarray]]
SinglePrediction = Union[jnp.ndarray, Mapping[str, jnp.ndarray]]

# In JAX, the term pytree refers to a tree-like structure built out of
# container-like Python objects.
# https://jax.readthedocs.io/en/latest/pytrees.html
PyTree = Any

# Flexible Params definition to support haiku as well as stax.
Params = PyTree

# Missing JAX types.
PRNGKey = jnp.ndarray

OptState = optax.OptState



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AgnosticFedAvg implementation.

Communication-Efficient Agnostic Federated Averaging
    Jae Ro, Mingqing Chen, Rajiv Mathews, Mehryar Mohri, Ananda Theertha Suresh
    https://arxiv.org/abs/2104.02748
"""

from typing import Any, Callable, List, Mapping, Optional, Sequence, Tuple

from fedjax.core import client_datasets
from fedjax.core import dataclasses
from fedjax.core import federated_algorithm
from fedjax.core import federated_data
from fedjax.core import for_each_client
from fedjax.core import optimizers
from fedjax.core import tree_util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey
from fedjax.core import util

import jax
import jax.numpy as jnp


def create_domain_metrics_for_each_client(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    num_domains: int,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
  """Creates for_each_client function for domain metrics calculation."""

  def client_init(shared_input, client_rng):
    step_state = {
        'params': shared_input['params'],
        'rng': client_rng,
        'domain_loss': jnp.zeros(num_domains),
        'domain_num': jnp.zeros(num_domains),
    }
    return step_state

  def client_step(step_state, batch):
    rng, use_rng = jax.random.split(step_state['rng'])
    example_mask = batch[client_datasets.EXAMPLE_MASK_KEY]
    example_loss = (
        per_example_loss(step_state['params'], batch, use_rng) * example_mask)
    domain_loss = jax.ops.segment_sum(example_loss, batch['domain_id'],
                                      num_domains)
    if regularizer is not None:
      domain_loss += regularizer(step_state['params'])
    domain_num = jax.ops.segment_sum(
        example_mask.astype(jnp.float32), batch['domain_id'], num_domains)
    next_step_state = {
        'params': step_state['params'],
        'rng': rng,
        'domain_loss': step_state['domain_loss'] + domain_loss,
        'domain_num': step_state['domain_num'] + domain_num
    }
    return next_step_state

  def client_final(shared_input, step_state):
    client_output = {
        'domain_loss': step_state['domain_loss'],
        'domain_num': step_state['domain_num'],
        'beta': jnp.sum(shared_input['alpha'] * step_state['domain_num'])
    }
    return client_output

  return for_each_client.for_each_client(client_init, client_step, client_final)


def create_scaled_loss(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    num_domains: int,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
  """Creates domain scaled loss function for agnostic client training."""

  def scaled_loss(params, batch, rng, alpha, beta):
    did = batch['domain_id']
    # Training batches produced by `ClientDataset.shuffle_repeat_batch` are not
    # padded so there's no need to worry about example masking.
    example_loss = per_example_loss(params, batch, rng)
    # \sum_{j \in b \cap \h{\sD}_i} \L(w, x_j, y_j)
    domain_sum_loss = jax.ops.segment_sum(example_loss, did, num_domains)
    loss = jnp.sum(alpha * domain_sum_loss) / beta
    if regularizer is not None:
      loss += regularizer(params)
    return loss

  return scaled_loss


def create_train_for_each_client(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    client_optimizer: optimizers.Optimizer,
    num_domains: int,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
  """Creates for_each_client for client training."""
  grad_fn = jax.grad(
      create_scaled_loss(per_example_loss, num_domains, regularizer))

  def client_init(shared_input, client_input):
    step_state = {
        'params': shared_input['params'],
        'alpha': shared_input['alpha'],
        'opt_state': client_optimizer.init(shared_input['params']),
        'rng': client_input['rng'],
        'beta': client_input['beta']
    }
    return step_state

  def client_step(step_state, batch):
    rng, use_rng = jax.random.split(step_state['rng'])
    grads = grad_fn(step_state['params'], batch, use_rng, step_state['alpha'],
                    step_state['beta'])
    opt_state, params = client_optimizer.apply(grads, step_state['opt_state'],
                                               step_state['params'])
    next_step_state = {
        **step_state,
        'params': params,
        'opt_state': opt_state,
        'rng': rng
    }
    return next_step_state

  def client_final(shared_input, step_state):
    delta_params = jax.tree_util.tree_multimap(lambda a, b: a - b,
                                               shared_input['params'],
                                               step_state['params'])
    return delta_params

  return for_each_client.for_each_client(client_init, client_step, client_final)


def update_domain_weights(domain_weights: jnp.ndarray, domain_loss: jnp.ndarray,
                          domain_learning_rate: float,
                          domain_algorithm: str) -> jnp.ndarray:
  """Updates domain weights following input algorithm."""
  if domain_algorithm == 'eg':
    new_domain_weights = domain_weights * jnp.exp(
        domain_learning_rate * domain_loss)
    new_domain_weights = jnp.maximum(new_domain_weights,
                                     jnp.zeros_like(new_domain_weights))
    return new_domain_weights / jnp.sum(new_domain_weights)
  elif domain_algorithm == 'none':
    return domain_weights
  else:
    raise ValueError(f'Unsupported domain algorithm {domain_algorithm!r}.')


@dataclasses.dataclass
class ServerState:
  """State of server for AgnosticFedAvg passed between rounds.

  Attributes:
    params: A pytree representing the server model parameters.
    opt_state: A pytree representing the server optimizer state.
    domain_weights: Weights per domain applied to weight in weighted average.
    domain_window: Sliding window keeping track of domain number of examples for
      the last window size rounds.
  """
  params: Params
  opt_state: optimizers.OptState
  domain_weights: jnp.ndarray
  domain_window: List[jnp.ndarray]


def agnostic_federated_averaging(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    client_optimizer: optimizers.Optimizer,
    server_optimizer: optimizers.Optimizer,
    client_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
    domain_batch_hparams: client_datasets.PaddedBatchHParams,
    init_domain_weights: Sequence[float],
    domain_learning_rate: float,
    domain_algorithm: str = 'eg',
    domain_window_size: int = 1,
    init_domain_window: Optional[Sequence[float]] = None,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None
) -> federated_algorithm.FederatedAlgorithm:
  """Builds agnostic federated averaging.

  Agnostic federated averaging requires input
  :class:`fedjax.core.client_datasets.ClientDataset` examples to contain
  a feature named "domain_id", which stores the integer domain id in
  [0, num_domains).
  For example, for Stack Overflow, each example post can be either a question or
  an answer, so there are two possible domain ids (question = 0; answer = 1).

  Args:
    per_example_loss: A function from (params, batch_example, rng) to a vector
      of loss values for each example in the batch. This is used in both the
      domain metrics computation and gradient descent training.
    client_optimizer: Optimizer for local client training.
    server_optimizer: Optimizer for server update.
    client_batch_hparams: Hyperparameters for client dataset for training.
    domain_batch_hparams: Hyperparameters for client dataset domain metrics
      calculation.
    init_domain_weights: Initial weights per domain that must sum to 1.
    domain_learning_rate: Learning rate for domain weight update.
    domain_algorithm: Algorithm used to update domain weights each round. One of
      'eg', 'none'.
    domain_window_size: Size of sliding window keeping track of number of
      examples per domain over multiple rounds.
    init_domain_window: Initial values for domain window. Defaults to ones.
    regularizer: Optional regularizer that only depends on params.

  Returns:
    FederatedAlgorithm.

  Raises:
    ValueError: If ``init_domain_weights`` does not sum to 1 or if
      ``init_domain_weights`` and ``init_domain_window`` are unequal lengths.
  """
  if abs(sum(init_domain_weights) - 1) > 1e-6:
    raise ValueError('init_domain_weights must sum to approximately 1.')

  if init_domain_window is None:
    init_domain_window = jnp.ones_like(init_domain_weights)

  if len(init_domain_weights) != len(init_domain_window):
    raise ValueError(
        f'init_domain_weights and init_domain_window must be equal lengths.'
        f' {len(init_domain_weights)} != {len(init_domain_window)}'
    )

  num_domains = len(init_domain_weights)
  domain_metrics_for_each_client = create_domain_metrics_for_each_client(
      per_example_loss, num_domains)
  train_for_each_client = create_train_for_each_client(per_example_loss,
                                                       client_optimizer,
                                                       num_domains, regularizer)

  def init(params: Params) -> ServerState:
    opt_state = server_optimizer.init(params)
    domain_weights = jnp.array(init_domain_weights)
    domain_window = [jnp.array(init_domain_window)] * domain_window_size
    return ServerState(params, opt_state, domain_weights, domain_window)

  def apply(
      server_state: ServerState,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]]
  ) -> Tuple[ServerState, Mapping[federated_data.ClientId, Any]]:
    # α
    alpha = server_state.domain_weights / jnp.mean(
        jnp.asarray(server_state.domain_window), axis=0)
    # First pass to calculate initial domain loss, domain num, and scaling
    # weight β for each client. This doesn't involve any aggregation at the
    # server, so this step and training can be a single round of communication.
    domain_batch_clients = [(cid, cds.padded_batch(domain_batch_hparams), crng)
                            for cid, cds, crng in clients]
    shared_input = {'params': server_state.params, 'alpha': alpha}
    # L^k, N^k, β^k
    client_domain_metrics = dict(
        domain_metrics_for_each_client(shared_input, domain_batch_clients))
    # Train for each client using scaling weights α and β.
    batch_clients = []
    for cid, cds, crng in clients:
      client_input = {'rng': crng, 'beta': client_domain_metrics[cid]['beta']}
      batch_clients.append(
          (cid, cds.shuffle_repeat_batch(client_batch_hparams), client_input))

    client_diagnostics = {}
    # Mean delta params across clients.
    delta_params_sum = tree_util.tree_zeros_like(server_state.params)
    weight_sum = 0.
    # w^k
    for cid, delta_params in train_for_each_client(shared_input, batch_clients):
      weight = client_domain_metrics[cid]['beta']
      delta_params_sum = tree_util.tree_add(
          delta_params_sum, tree_util.tree_weight(delta_params, weight))
      weight_sum += weight
      client_diagnostics[cid] = {
          'delta_l2_norm': tree_util.tree_l2_norm(delta_params)
      }
    mean_delta_params = tree_util.tree_inverse_weight(delta_params_sum,
                                                      weight_sum)
    # Sum domain metrics across clients.
    sum_domain_loss = tree_util.tree_sum(
        d['domain_loss'] for d in client_domain_metrics.values())
    sum_domain_num = tree_util.tree_sum(
        d['domain_num'] for d in client_domain_metrics.values())
    server_state = server_update(server_state, mean_delta_params,
                                 sum_domain_loss, sum_domain_num)
    return server_state, client_diagnostics

  def server_update(server_state, mean_delta_params, sum_domain_loss,
                    sum_domain_num):
    opt_state, params = server_optimizer.apply(mean_delta_params,
                                               server_state.opt_state,
                                               server_state.params)
    mean_domain_loss = util.safe_div(sum_domain_loss, sum_domain_num)
    domain_weights = update_domain_weights(server_state.domain_weights,
                                           mean_domain_loss,
                                           domain_learning_rate,
                                           domain_algorithm)
    domain_window = server_state.domain_window[1:] + [sum_domain_num]
    return ServerState(params, opt_state, domain_weights, domain_window)

  return federated_algorithm.FederatedAlgorithm(init, apply)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for in_memory_federated_data."""

from absl.testing import absltest
from fedjax.core import client_datasets
from fedjax.core import in_memory_federated_data
import numpy as np
import numpy.testing as npt


class InMemoryFederatedDataTest(absltest.TestCase):

  def test_in_memory_data_test(self):
    client_a_data = {
        'x': np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
        'y': np.array([7, 8])
    }
    client_b_data = {'x': np.array([[9.0, 10.0, 11.0]]), 'y': np.array([12])}
    client_to_data_mapping = {'a': client_a_data, 'b': client_b_data}

    federated_data = in_memory_federated_data.InMemoryFederatedData(
        client_to_data_mapping)

    with self.subTest('num_clients'):
      self.assertEqual(federated_data.num_clients(), 2)

    with self.subTest('client_ids'):
      self.assertEqual(list(federated_data.client_ids()), ['a', 'b'])

    with self.subTest('client_sizes'):
      self.assertEqual(
          list(federated_data.client_sizes()), [('a', 2), ('b', 1)])

    with self.subTest('client_size'):
      self.assertEqual(federated_data.client_size('a'), 2)

    with self.subTest('get_client'):
      self.assertEqual(
          federated_data.get_client('a').all_examples(), client_a_data)

    with self.subTest('get_clients'):
      clients_data = list(federated_data.get_clients(['a']))
      self.assertLen(clients_data, 1)
      client_id, client = clients_data[0]
      self.assertEqual(client_id, 'a')
      self.assertEqual(client.all_examples(), client_a_data)

    with self.subTest('clients'):
      all_client_datasets = list(federated_data.clients())
      self.assertEqual(
          [(key, value.all_examples()) for (key, value) in all_client_datasets],
          [(key, value) for key, value in client_to_data_mapping.items()])

    with self.subTest('preprocess_client'):

      def preprocess_client(client_id, examples):
        del client_id
        examples['x'] = 2 * examples['x']
        return examples

      new_federeated_data = federated_data.preprocess_client(preprocess_client)
      new_client_a_data = {
          'x': np.array([[2.0, 4.0, 6.0], [8.0, 10.0, 12.0]]),
          'y': np.array([7, 8])
      }
      npt.assert_array_equal(
          new_federeated_data.get_client('a').all_examples()['x'],
          new_client_a_data['x'])
      npt.assert_array_equal(
          new_federeated_data.get_client('a').all_examples()['y'],
          new_client_a_data['y'])

    with self.subTest('preprocess_batch'):

      def preprocess_batch(examples):
        examples['x'] = 2 * examples['x']
        return examples

      new_federeated_data = federated_data.preprocess_batch(preprocess_batch)
      new_client_a_data = {
          'x': np.array([[2.0, 4.0, 6.0], [8.0, 10.0, 12.0]]),
          'y': np.array([7, 8])
      }
      batched_client_a_data = list(
          new_federeated_data.get_client('a').batch(
              hparams=client_datasets.BatchHParams(2)))[0]
      npt.assert_array_equal(batched_client_a_data['x'], new_client_a_data['x'])
      npt.assert_array_equal(batched_client_a_data['y'], new_client_a_data['y'])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Preprocessing and batching operations over client datasets.

**Column based representation**

The examples in a client dataset can be viewed as a table, where the rows are
the individual examples, and the columns are the features (labels are viewed as
a feature in this context).

We use a column based representation when loading a dataset into memory.

- Each column is a NumPy array ``x`` of rank at least 1, where ``x[i, ...]`` is
  the value of this feature for the ``i``-th example.
- The complete set of examples is a dict-like object, from ``str`` feature
  names, to the corresponding column values.

Traditionally, a row based representation is used for representing the entire
dataset, and a column based representation is used for a single batch. In the
context of federated learning, an individual client dataset is small enough to
easily fit into memory so the same representation is used for the entire dataset
and a batch.

**Preprocessor**

Preprocessing on a batch of examples can be easily done via a chain of
functions. A ``Preprocessor`` object holds the chain of functions, and applies
the transformation on a batch of examples.

**ClientDataset: examples + preprocessor**

A :class:`~fedjax.ClientDataset` is simply some examples in the column based
representation, accompanied by a Preprocessor.
Its :meth:`~fedjax.ClientDataset.batch` method produces batches of examples in a
sequential order, suitable for evaluation.
Its :meth:`~fedjax.ClientDataset.shuffle_repeat_batch` method adds shuffling and
repeating, making it suitable for training.
"""

import collections
import itertools
from typing import Any, Callable, Iterable, Iterator, Mapping, Optional

from fedjax.core import dataclasses
import numpy as np

# The same column based representation for examples in the entire client
# dataset, or in a single batch.
Examples = Mapping[str, np.ndarray]


class BatchPreprocessor:
  """A chain of preprocessing functions on batched examples.

  BatchPreprocessor holds a chain of preprocessing functions, and applies them
  in order on batched examples. Each individual preprocessing function operates
  over multiple examples, instead of just 1 example. For example, ::

    preprocessor = BatchPreprocessor([
      # Flattens `pixels`.
      lambda x: {**x, 'pixels': x['pixels'].reshape([-1, 28 * 28])},
      # Introduce `binary_label`.
      lambda x: {**x, 'binary_label': x['label'] % 2},
    ])
    fake_emnist = {
      'pixels': np.random.uniform(size=(10, 28, 28)),
      'label': np.random.randint(10, size=(10,))
    }
    preprocessor(fake_emnist)
    # Produces a dict of [10, 28*28] "pixels", [10,] "label" and "binary_label".

  Given a BatchPreprocessor, a new BatchPreprocessor can be created with an
  additional preprocessing function appended to the chain, ::

    # Continuing from the previous example.
    new_preprocessor = preprocessor.append(
      lambda x: {**x, 'sum_pixels': np.sum(x['pixels'], axis=1)})
    new_preprocessor(fake_emnist)
    # Produces a dict of [10, 28*28] "pixels", [10,] "sum_pixels", "label" and
    # "binary_label".

  The main difference of this preprocessor and
  :class:`fedjax.ClientPreprocessor` is that :class:`fedjax.ClientPreprocessor`
  also takes ``client_id`` as input. Because of the identical representation
  between batched examples and all examples in a client dataset, certain
  preprocessing can be done with either BatchPreprocessor or ClientPreprocessor.

  **Examples of preprocessing possible at either the client dataset level, or
  the batch level**

  Such preprocessing is deterministic, and strictly per-example.

  - Casting a feature from `int8` to `float32`.
  - Adding a new feature derived from existing features.
  - Remove a feature (although the better place to do so is at the dataset
    level).

  A simple rule for deciding where to carry out the preprocessing in this case
  is the following,

  - Does this make batching cheaper (e.g. removing features)? If so, do it at
    the dataset level.
  - Otherwise, do it at the batch level.

  Assuming preprocessing time is linear in the number of examples, preprocessing
  at the batch level has the benefit of evenly distributing host compute work,
  which may overlap better with asynchronous JAX compute work on GPU/TPU.

  **Examples of preprocessing only possible at the batch level**

  - Data augmentation (e.g. random cropping).
  - Padding at the batch size dimension.

  **Examples of preprocessing only possible at the dataset level**

  - Those that require knowing the client id.
  - Capping the number of examples.
  - Altering what it means to be an example: e.g. in certain language model
    setups, sentences are concatenated and then split into equal sized chunks.
  """

  def __init__(self, fns: Iterable[Callable[[Examples], Examples]] = ()):
    self._fns = tuple(fns)

  def __call__(self, examples: Examples) -> Examples:
    if not self._fns:
      return examples
    # Make a copy to guard against fns that modify their input.
    out = dict(examples)
    for f in self._fns:
      out = f(out)
    assert_consistent_rows(out)
    return out

  def append(self, fn: Callable[[Examples], Examples]) -> 'BatchPreprocessor':
    """Creates a new BatchPreprocessor with fn added to the end."""
    return BatchPreprocessor(self._fns + (fn,))

  def __str__(self) -> str:
    return f'BatchPreprocessor({self._fns})'

  def __repr__(self) -> str:
    return str(self)


# A common default preprocessor that does nothing.
NoOpBatchPreprocessor = BatchPreprocessor()


# A special feature name for padded batches.
EXAMPLE_MASK_KEY = '__mask__'


# We group hyperparams to batching functions so that,
#
# 1.  They can be easily passed around as a single object, instead of a list of
#     arguments.
# 2.  The list of argument is only defined, documented, and assigned default
#     values once (in the hyperparams class).
#
# This setup is more convenient for library code where the caller specifies such
# a hyperparams object, but less so for users directly calling the actual
# batching function (e.g. compare `padded_batch(batch_size=3)` vs
# `padded_batch(PaddedBatchHParams(batch_size=3))`. Therefore, all our batching
# functions support two ways of invocation:
# -   They can be passed in a hyperparams object.
# -   They can also be passed in a list of keyword arguments, which will be used
#     to construct the hyperparams object.
#
# See ClientDataset.padded_batch() for an example.


@dataclasses.dataclass
class PaddedBatchHParams:
  """See :meth:`ClientDataset.padded_batch`.

  Attributes:
    batch_size: Desired batch size.
    num_batch_size_buckets: Number of batch size buckets for the final batch.
  """
  batch_size: int
  num_batch_size_buckets: int = 1


@dataclasses.dataclass
class ShuffleRepeatBatchHParams:
  """See :meth`:ClientDataset.shuffle_repeat_batch`.

  Attributes:
    batch_size: Desired batch size.
    num_epochs: Optional number of passes to iterate over the dataset.
    num_steps: Optional number of batches to produce.
    drop_remainder: Whether to drop a trailing batch smaller than ``batch_size``.
    seed: Optional random number generator seed.
  """
  batch_size: int
  num_epochs: Optional[int] = 1
  num_steps: Optional[int] = None
  drop_remainder: bool = False
  seed: Optional[int] = None


@dataclasses.dataclass
class BatchHParams:
  """See :meth:`ClientDataset.batch`.

  Attributes:
    batch_size: Desired batch size.
    drop_remainder: Whether to drop the final batch when it contains fewer than
      ``batch_size`` examples.
  """
  batch_size: int
  drop_remainder: bool = False


class ClientDataset:
  """In memory client dataset backed by numpy ndarrays.

  Custom preprocessing on batches can be added via a preprocessor. A
  ClientDataset is stored as the unpreprocessed :attr:`raw_examples`, along with
  its preprocessor.

  - To access batches, use one of the batching functions (e.g.
    :meth:`shuffle_repeat_batch` for training, :meth:`padded_batch`
    for evaluation).
  - To access a small number of preprocessed examples (e.g. for exploration),
    use slicing + :meth:`all_examples`.

  This is only intended for efficient access to small datasets that fit in
  memory.
  """

  def __init__(self,
               raw_examples: Examples,
               preprocessor: BatchPreprocessor = NoOpBatchPreprocessor):
    assert_consistent_rows(raw_examples)
    self.raw_examples = raw_examples
    self.preprocessor = preprocessor

  def __len__(self) -> int:
    """Returns the number of raw examples in this dataset."""
    return num_examples(self.raw_examples, validate=False)

  def __getitem__(self, index: slice) -> 'ClientDataset':
    """Returns a new ClientDataset with sliced raw examples."""
    if not isinstance(index, slice):
      raise ValueError(f'Only slicing is supported, got index {index!r}')
    return ClientDataset(
        slice_examples(self.raw_examples, index), self.preprocessor)

  def all_examples(self) -> Examples:
    """Returns the result of feeding all raw examples through the preprocessor.

    This is mostly intended for interactive exploration of a small subset of a
    client dataset. For example, to see the first 4 examples in a client
    dataset, ::

      dataset = ClientDataset(my_raw_examples, my_preprocessor)
      dataset[:4].all_examples()

    Returns:
      Preprocessed examples from all the raw examples in this client dataset.
    """
    return self.preprocessor(self.raw_examples)

  def padded_batch(self,
                   hparams: Optional[PaddedBatchHParams] = None,
                   **kwargs) -> Iterable[Examples]:
    """Produces preprocessed padded batches in a fixed sequential order.

    This function can be invoked in 2 ways:

    1.  Using a hyperparams object. This is the recommended way in library code.
        Example::

          def a_library_function(client_dataset, hparams):
            for batch in client_dataset.padded_batch(hparams):
              ...

    2.  Using keyword arguments. The keyword arguments are used to construct
        a new hyperparams object, or override an existing one. For example, ::

          client_dataset.padded_batch(batch_size=2)
          # Overrides the default num_batch_size_buckets value.
          client_dataset.padded_batch(hparams, num_batch_size_buckets=2)

    When the number of examples in the dataset is not a multiple of
    ``batch_size``, the final batch may be smaller than ``batch_size``.
    This may lead to a large number of JIT recompilations. This can be
    circumvented by padding the final batch to a small number of fixed sizes
    controlled by ``num_batch_size_buckets``.

    All batches contain an extra bool feature keyed by ``EXAMPLE_MASK_KEY``.
    ``batch[EXAMPLE_MASK_KEY][i]`` tells us whether the ``i``-th example in this
    batch is an actual example (``batch[EXAMPLE_MASK_KEY][i] == True``), or a
    padding example (``batch[EXAMPLE_MASK_KEY][i] == False``).

    We repeatedly halve the batch size up to ``num_batch_size_buckets-1`` times,
    until we find the smallest one that is also >= the size of the final batch.
    Therefore if ``batch_size < 2^num_batch_size_buckets``, fewer bucket sizes
    will be actually used.

    Args:
      hparams: Batching hyperparameters.
      **kwargs: Keyword arguments for constructing/overriding hparams.

    Returns:
      An iterable object that can be iterated over multiple times.
    """
    if hparams is None:
      hparams = PaddedBatchHParams(**kwargs)
    elif kwargs:
      hparams = hparams.replace(**kwargs)
    return PaddedBatchView(self, hparams)

  def shuffle_repeat_batch(self,
                           hparams: Optional[ShuffleRepeatBatchHParams] = None,
                           **kwargs) -> Iterable[Examples]:
    """Produces preprocessed batches in a shuffled and repeated order.

    This function can be invoked in 2 ways:

    1.  Using a hyperparams object. This is the recommended way in library code.
        Example::

          def a_library_function(client_dataset, hparams):
            for batch in client_dataset.shuffle_repeat_batch(hparams):
              ...

    2.  Using keyword arguments. The keyword arguments are used to construct
        a new hyperparams object, or override an existing one. For example, ::

          client_dataset.shuffle_repeat_batch(batch_size=2)
          # Overrides the default num_epochs value.
          client_dataset.shuffle_repeat_batch(hparams, num_epochs=2)

    Shuffling is done without replacement, therefore for a dataset of N
    examples, the first ``ceil(N/batch_size)`` batches are guarranteed to cover
    the entire dataset.

    By default the iteration stops after the first epoch. The number of batches
    produced from the iteration can be controlled by the
    ``(num_epochs, num_steps, drop_remainder)`` combination:

    - If both ``num_epochs`` and ``num_steps`` are None, the shuffle-repeat
      process continues forever.
    - If ``num_epochs`` is set and ``num_steps`` is None, as few batches as needed
      to go over the dataset this many passes are produced. Further,

        - If ``drop_remainder`` is False (the default), the final batch is
          filled with additionally sampled examples to contain ``batch_size``
          examples.
        - If ``drop_remainder`` is True, the final batch is dropped if it
          contains fewer than ``batch_size`` examples. This may result in
          examples being skipped when ``num_epochs=1``.
    - If ``num_steps`` is set and ``num_steps`` is None, exactly this many batches
      are produced. ``drop_remainder`` has no effect in this case.
    - If both ``num_epochs`` and ``num_steps`` are set, the fewer number of
      batches between the two conditions are produced.

    If reproducible iteration order is desired, a fixed ``seed`` can be used. When
    ``seed`` is None, repeated iteration over the same object may produce batches
    in a different order.

    Unlike ``batch()`` or ``padded_batch()``, batches from ``shuffle_repeat_batch()``
    always contain exactly ``batch_size`` examples. Also unlike TensorFlow, that
    holds even when ``drop_remainder=False``.

    Args:
      hparams: Batching hyperparamters.
      **kwargs: Keyword arguments for constructing/overriding hparams.

    Returns:
      An iterable object that can be iterated over multiple times.
    """
    if hparams is None:
      hparams = ShuffleRepeatBatchHParams(**kwargs)
    elif kwargs:
      hparams = hparams.replace(**kwargs)
    return ShuffleRepeatBatchView(self, hparams)

  def batch(self,
            hparams: Optional[BatchHParams] = None,
            **kwargs) -> Iterable[Examples]:
    """Produces preprocessed batches in a fixed sequential order.

    The final batch may contain fewer than ``batch_size`` examples. If used
    directly, that may result in a large number of JIT recompilations. Therefore
    we recommended using ``padded_batch`` instead in most scenarios.

    This function can be invoked in 2 ways:

    1.  Using a hyperparams object. This is the recommended way in library code
        if you have to use batch (prefer padded_batch() if possible). Example::

          def a_library_function(client_dataset, hparams):
            for batch in client_dataset.batch(hparams):
              ...

    2.  Using keyword arguments. The keyword arguments are used to construct
        a new hyperparams object, or override an existing one. For example, ::

          client_dataset.batch(batch_size=2)
          # Overrides the default drop_remainder value.
          client_dataset.batch(hparams, drop_remainder=True)

    Args:
      hparams: Batching hyperparameters.
      **kwargs: Keyword arguments for constructing/overriding hparams.

    Returns:
      An iterable object that can be iterated over multiple times.
    """
    if hparams is None:
      hparams = BatchHParams(**kwargs)
    elif kwargs:
      hparams = hparams.replace(**kwargs)
    return BatchView(self, hparams)


class PaddedBatchView:
  """View of ordered padded batches of a ClientDataset.

  See ClientDataset.padded_batch() for the expected behavior.
  """

  def __init__(self, client_dataset: ClientDataset,
               hparams: PaddedBatchHParams):
    self._client_dataset = client_dataset
    self._data_size = len(client_dataset)
    self._batch_size = hparams.batch_size
    self._final_batch_size = _pick_final_batch_size(
        self._data_size, self._batch_size, hparams.num_batch_size_buckets)

  def __iter__(self) -> Iterator[Examples]:
    # Mask for full batches.
    full_mask = np.ones([self._batch_size], dtype=np.bool_)
    for start in range(0, self._data_size, self._batch_size):
      stop = start + self._batch_size
      sliced = slice_examples(self._client_dataset.raw_examples,
                              slice(start, stop))
      processed = self._client_dataset.preprocessor(sliced)
      if stop <= self._data_size:
        yield {**processed, EXAMPLE_MASK_KEY: full_mask}
      else:
        yield pad_examples(processed, self._final_batch_size)


def _pick_final_batch_size(data_size: int, batch_size: int,
                           num_batch_size_buckets: int) -> int:
  """Picks the final batch size for a given dataset size."""
  # Determine the batch size for the final batch.
  final_batch_size = data_size % batch_size
  if final_batch_size == 0:
    # No padding necessary.
    return batch_size
  # final_batch_size in [1, batch_size)
  high, low, n = batch_size, batch_size // 2, 1
  # Find low < final_batch_size <= high
  while low >= final_batch_size and n < num_batch_size_buckets:
    high, low, n = low, low // 2, n + 1
  return high


class ShuffleRepeatBatchView:
  """View of shuffled and repeated batches of a ClientDataset.

  See ClientDataset.shuffle_repeat_batch() for the expected behavior.
  """

  def __init__(self, client_dataset: ClientDataset,
               hparams: ShuffleRepeatBatchHParams):
    self._client_dataset = client_dataset
    self._data_size = len(client_dataset)
    self._batch_size = hparams.batch_size
    if hparams.num_epochs is not None:
      if hparams.drop_remainder:
        self._num_steps = (
            self._data_size * hparams.num_epochs // hparams.batch_size)
      else:
        self._num_steps = (
            (self._data_size * hparams.num_epochs + hparams.batch_size - 1) //
            hparams.batch_size)
      if hparams.num_steps is not None:
        self._num_steps = min(hparams.num_steps, self._num_steps)
    elif hparams.num_steps is not None:
      self._num_steps = hparams.num_steps
    else:
      self._num_steps = None
    self._seed = hparams.seed

  def __iter__(self) -> Iterator[Examples]:
    buf = np.arange(self._data_size, dtype=np.int32)
    buf_size = buf.shape[0]
    # Start of unused portion of buf. We start with no unused values because we
    # haven't shuffled yet.
    i = buf_size
    num_steps = 0
    desired_num_steps = self._num_steps
    rng = np.random.RandomState(self._seed)
    while desired_num_steps is None or num_steps < desired_num_steps:
      # Find example indices of next batch.
      indices = np.zeros((self._batch_size,), dtype=np.int32)
      desired_size = indices.size
      filled = 0
      while filled < desired_size:
        available = buf_size - i
        if available == 0:
          rng.shuffle(buf)
          i = 0
          available = buf_size
        used = min(available, desired_size - filled)
        indices[filled:filled + used] = buf[i:i + used]
        i += used
        filled += used
      # Produce next batch.
      sliced = {
          k: v[indices] for k, v in self._client_dataset.raw_examples.items()
      }
      yield self._client_dataset.preprocessor(sliced)
      num_steps += 1


class BatchView:
  """View of ordered batches of a ClientDataset.

  See ClientDataset.batch() for the expected behavior.
  """

  def __init__(self, client_dataset: ClientDataset, hparams: BatchHParams):
    self._client_dataset = client_dataset
    self._batch_size = hparams.batch_size
    self._drop_remainder = hparams.drop_remainder
    self._data_size = len(client_dataset)

  def __iter__(self) -> Iterator[Examples]:
    for start in range(0, self._data_size, self._batch_size):
      stop = start + self._batch_size
      sliced = slice_examples(self._client_dataset.raw_examples,
                              slice(start, stop))
      processed = self._client_dataset.preprocessor(sliced)
      if not self._drop_remainder or stop <= self._data_size:
        yield processed


def padded_batch_client_datasets(datasets: Iterable[ClientDataset],
                                 hparams: Optional[PaddedBatchHParams] = None,
                                 **kwargs) -> Iterator[Examples]:
  """Batches examples from multiple client datasets.

  This is useful when we want to evaluate on the combined dataset consisting of
  multiple client datasets. Unlike batching each client dataset individually, we
  can reduce the number of batches smaller than ``batch_size``.

  This function can be invoked in 2 ways:

  1.  Using a hyperparams object. This is the recommended way in library code.
      Example::

        def a_library_function(datasets, hparams):
          for batch in padded_batch_client_datasets(datasets, hparams):
            ...

  2.  Using keyword arguments. The keyword arguments are used to construct
      a new hyperparams object, or override an existing one. For example, ::

        padded_batch_client_datasets(datasets, hparams)
        # Overrides the default num_batch_size_buckets value.
        padded_batch_client_datasets(datasets, hparams, num_batch_size_buckets=2)

  Args:
    datasets: ClientDatasets to be batched. All ClientDatasets must have the
      same Preprocessor object attached.
    hparams: Batching hyperparams like those in :meth:`ClientDataset.padded_batch`.
    **kwargs: Keyword arguments for constructing/overriding hparams.

  Yields:
    Batches of examples. The final batch might be padded. All batches contain
    a bool feature keyed by `EXAMPLE_MASK_KEY`.

  Raises:
    ValueError: If any 2 client datasets have different Preprocessors.
    ValueError: If any 2 client datasets have different features.
  """
  if hparams is None:
    hparams = PaddedBatchHParams(**kwargs)
  elif kwargs:
    hparams = hparams.replace(**kwargs)
  preprocessor = None
  features = None
  # Pieces of examples whose total size is < batch_size
  buf = []
  # Total size of examples in buf.
  buf_size = 0
  # Mask for full batches.
  full_mask = np.ones([hparams.batch_size], dtype=np.bool_)
  # Invariant: buf_size < batch_size.
  for dataset in datasets:
    if preprocessor is None:
      preprocessor = dataset.preprocessor
    elif dataset.preprocessor is not preprocessor:
      raise ValueError(
          'client_datasets should have the identical Preprocessor object, '
          f'got {preprocessor} vs {dataset.preprocessor}')
    if features is None:
      features = set(dataset.raw_examples)
    elif features != set(dataset.raw_examples):
      raise ValueError('client_datasets should have identical features, '
                       f'got {features} vs {list(dataset.raw_examples)}')
    size = len(dataset)
    examples = dataset.raw_examples
    if buf_size + size < hparams.batch_size:
      buf.append(examples)
      buf_size += size
      continue
    if buf:
      # Emit what's in buf.
      start = hparams.batch_size - buf_size
      buf.append(slice_examples(examples, slice(start)))
      yield attach_mask(preprocessor(concat_examples(buf)), full_mask)
      buf.clear()
      buf_size = 0
    else:
      start = 0
    # Emit batches in examples.
    while start + hparams.batch_size < size:
      yield attach_mask(
          preprocessor(
              slice_examples(examples, slice(start,
                                             start + hparams.batch_size))),
          full_mask)
      start += hparams.batch_size
    # Buffer remaining.
    if start < size:
      buf.append(slice_examples(examples, slice(start, size)))
      buf_size += size - start
  if buf:
    final_examples = preprocessor(concat_examples(buf))
    final_batch_size = _pick_final_batch_size(buf_size, hparams.batch_size,
                                              hparams.num_batch_size_buckets)
    yield pad_examples(final_examples, final_batch_size)


def buffered_shuffle(source: Iterable[Any], buffer_size: int,
                     rng: np.random.RandomState) -> Iterator[Any]:
  """Shuffles an iterable via buffered shuffling."""
  it = iter(source)
  buf = list(itertools.islice(it, buffer_size))
  rng.shuffle(buf)
  for i in it:
    r, buf[0] = buf[0], i
    swap = rng.randint(buffer_size)
    if swap < buffer_size - 1:
      buf[swap], buf[0] = buf[0], buf[swap]
    yield r
  for i in buf:
    yield i


def buffered_shuffle_batch_client_datasets(
    datasets: Iterable[ClientDataset], batch_size: int, buffer_size: int,
    rng: np.random.RandomState) -> Iterator[Examples]:
  """Shuffles and batches examples from multiple client datasets.

  This just makes 1 pass over the examples. To achieve repeated iterations,
  create an infinite shuffled stream of datasets first (e.g. using
  buffered_shuffle()).

  Args:
    datasets: ClientDatasets to be batched. All ClientDatasets must have the
      same Preprocessor object attached.
    batch_size: Desired batch size.
    buffer_size: Number of examples to buffer during shuffling.
    rng: Source of randomness.

  Yields:
    Batches of examples. For a finite stream of datasets, the final batch might
    be smaller than ``batch_size``.

  Raises:
    ValueError: If any 2 client datasets have different Preprocessors.
    ValueError: If any 2 client datasets have different features.
  """

  def gen_items():
    """Generates the processor, then pairs of (examples, index)."""
    preprocessor = None
    features = None
    for dataset in datasets:
      if preprocessor is None:
        preprocessor = dataset.preprocessor
        yield preprocessor
      elif dataset.preprocessor is not preprocessor:
        raise ValueError(
            'client_datasets should have the identical Preprocessor object, '
            f'got {preprocessor} vs {dataset.preprocessor}')
      if features is None:
        features = set(dataset.raw_examples)
      elif features != set(dataset.raw_examples):
        raise ValueError('client_datasets should have identical features, '
                         f'got {features} vs {list(dataset.raw_examples)}')
      for i in range(len(dataset)):
        yield (dataset.raw_examples, i)

  it = gen_items()
  try:
    preprocessor = next(it)
  except StopIteration:
    return
  buf = []
  for item in buffered_shuffle(it, buffer_size, rng):
    buf.append(item)
    if len(buf) == batch_size:
      yield preprocessor(
          concat_examples([slice_examples(e, slice(i, i + 1)) for e, i in buf]))
      buf.clear()
  if buf:
    yield preprocessor(
        concat_examples([slice_examples(e, slice(i, i + 1)) for e, i in buf]))


def assert_consistent_rows(examples: Examples):
  """Asserts `examples` have consistent row sizes (the number of examples)."""
  sizes = {k: v.shape[0] for k, v in examples.items()}
  if not sizes:
    raise ValueError('No features in examples')
  it = iter(sizes.items())
  name, size = next(it)
  for k, v in it:
    if v != size:
      raise ValueError(
          f'Feature {name} has {size} rows, but feature {k} has {v} rows')


def num_examples(examples: Examples, validate: bool = True) -> int:
  if validate:
    assert_consistent_rows(examples)
  return len(next(iter(examples.values())))


def slice_examples(examples: Examples, index: slice) -> Examples:
  return {k: v[index] for k, v in examples.items()}


def concat_examples(many_examples: Iterable[Examples]) -> Examples:
  combined = collections.defaultdict(list)
  for examples in many_examples:
    for k, v in examples.items():
      combined[k].append(v)
  return {k: np.concatenate(v, axis=0) for k, v in combined.items()}


def attach_mask(examples: Examples, mask: np.ndarray) -> Examples:
  if EXAMPLE_MASK_KEY in examples:
    raise ValueError(
        f'mask key {EXAMPLE_MASK_KEY!r} is already present in examples {examples}'
    )
  return {**examples, EXAMPLE_MASK_KEY: mask}


def pad_examples(examples: Examples, size: int) -> Examples:
  """Pad examples to a fixed number of rows with 0 values.

  Args:
    examples: Examples to pad.
    size: The desired number of rows (i.e. the leading dimension size).

  Returns:
    Padded examples.

  Raises:
    ValueError: invalid inputs.
  """
  if EXAMPLE_MASK_KEY in examples:
    raise ValueError(
        f'mask key {EXAMPLE_MASK_KEY!r} is already present in examples {examples}'
    )
  current_size = num_examples(examples)
  if current_size > size:
    raise ValueError(f'Cannot pad {current_size} examples to size {size}')
  result = {EXAMPLE_MASK_KEY: np.arange(size) < current_size}
  for k, v in examples.items():
    padded = np.zeros((size,) + v.shape[1:], v.dtype)
    padded[:current_size] = v
    result[k] = padded
  return result



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated hypothesis-based clustering implementation.

Three Approaches for Personalization with Applications to Federated Learning
    Yishay Mansour, Mehryar Mohri, Jae Ro, Ananda Theertha Suresh
    https://arxiv.org/abs/2002.10619
"""
import functools
from typing import Any, Callable, Dict, Iterable, Iterator, List, Mapping, Optional, Sequence, Tuple

from fedjax.core import client_datasets
from fedjax.core import dataclasses
from fedjax.core import federated_algorithm
from fedjax.core import federated_data
from fedjax.core import for_each_client
from fedjax.core import models
from fedjax.core import optimizers
from fedjax.core import tree_util
from fedjax.core.typing import BatchExample
from fedjax.core.typing import OptState
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey

import jax
import jax.numpy as jnp


@dataclasses.dataclass
class ServerState:
  """State of server for HypCluster passed between rounds.

  Attributes:
    cluster_params: A list of pytrees representing the server model parameters
      per cluster.
    opt_states: A list of pytrees representing the server optimizer state per
      cluster.
  """
  cluster_params: List[Params]
  opt_states: List[OptState]


def hyp_cluster(
    per_example_loss: Callable[[Params, BatchExample, PRNGKey], jnp.ndarray],
    client_optimizer: optimizers.Optimizer,
    server_optimizer: optimizers.Optimizer,
    maximization_batch_hparams: client_datasets.PaddedBatchHParams,
    expectation_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
    regularizer: Optional[Callable[[Params], jnp.ndarray]] = None
) -> federated_algorithm.FederatedAlgorithm:
  """Federated hypothesis-based clustering algorithm.

  Args:
    per_example_loss: A function from (params, batch, rng) to a vector of per
      example loss values.
    client_optimizer: Client side optimizer.
    server_optimizer: Server side optimizer.
    maximization_batch_hparams: Batching hyperparameters for the maximization
      step.
    expectation_batch_hparams: Batching hyperparameters for the expectation
      step.
    regularizer: Optional regularizer.

  Returns:
    A FederatedAlgorithm. A notable difference from other common
    FederatedAlgorithms such as fed_avg is that `init()` takes a list of
    `cluster_params`, which can be obtained using `random_init()`,
    `ModelKMeansInitializer`, or `kmeans_init()` in this module.
  """

  def init(cluster_params: List[Params]) -> ServerState:
    return ServerState(
        cluster_params,
        [server_optimizer.init(params) for params in cluster_params])

  # Creating these objects outside apply() can speed up repeated apply() calls.
  evaluator = models.AverageLossEvaluator(per_example_loss, regularizer)
  trainer = ClientDeltaTrainer(
      models.grad(per_example_loss, regularizer), client_optimizer)

  def apply(
      server_state: ServerState,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]]
  ) -> Tuple[ServerState, Mapping[federated_data.ClientId, Any]]:
    # Split RNGs for the 2 steps below.
    client_rngs = [jax.random.split(rng) for _, _, rng in clients]

    client_cluster_ids = maximization_step(
        evaluator=evaluator,
        cluster_params=server_state.cluster_params,
        clients=[(client_id, dataset, rng[0])
                 for (client_id, dataset, _), rng in zip(clients, client_rngs)],
        batch_hparams=maximization_batch_hparams)

    cluster_delta_params = expectation_step(
        trainer=trainer,
        cluster_params=server_state.cluster_params,
        client_cluster_ids=client_cluster_ids,
        clients=[(client_id, dataset, rng[1])
                 for (client_id, dataset, _), rng in zip(clients, client_rngs)],
        batch_hparams=expectation_batch_hparams)

    # Apply delta params for each cluster.
    cluster_params = []
    opt_states = []
    for delta_params, opt_state, params in zip(cluster_delta_params,
                                               server_state.opt_states,
                                               server_state.cluster_params):
      if delta_params is None:
        # No examples were observed for this cluster.
        next_opt_state, next_params = opt_state, params
      else:
        next_opt_state, next_params = server_optimizer.apply(
            delta_params, opt_state, params)
      cluster_params.append(next_params)
      opt_states.append(next_opt_state)

    # TODO(wuke): Other client diagnostics.
    client_diagnostics = {}
    for client_id, cluster_id in client_cluster_ids.items():
      client_diagnostics[client_id] = {'cluster_id': cluster_id}
    return ServerState(cluster_params, opt_states), client_diagnostics

  return federated_algorithm.FederatedAlgorithm(init, apply)


class _BaseClientTrainer:
  """Trains each client dataset, starting from either global params, or per client params."""

  def __init__(self, grad_fn: Callable[[Params, BatchExample, PRNGKey],
                                       jnp.ndarray],
               client_optimizer: optimizers.Optimizer, return_delta):
    # params can be passed in 2 ways:
    # -   As `shared_input`: All clients are trained starting from the same
    #     params.
    # -   As `client_input`: Each client is trained starting from per client
    #     params.
    def client_init(shared_input, client_input):
      if shared_input is not None:
        init_params = shared_input
        rng = client_input
      else:
        rng, init_params = client_input
      params = init_params
      opt_state = client_optimizer.init(params)
      return rng, params, opt_state, init_params

    def client_step(state, batch):
      rng, params, opt_state, init_params = state
      rng, use_rng = jax.random.split(rng)
      grads = grad_fn(params, batch, use_rng)
      opt_state, params = client_optimizer.apply(grads, opt_state, params)
      return rng, params, opt_state, init_params

    def client_final(shared_input, state):
      del shared_input
      _, params, _, init_params = state
      if return_delta:
        delta_params = jax.tree_util.tree_multimap(jnp.subtract, init_params,
                                                   params)
        return delta_params
      return params

    self._train_each_client = for_each_client.for_each_client(
        client_init, client_step, client_final)
    self._return_delta = return_delta

  def returns_delta(self):
    return self._return_delta

  def train_global_params(
      self, params: Params, clients: Iterable[Tuple[federated_data.ClientId,
                                                    Iterable[BatchExample],
                                                    PRNGKey]]
  ) -> Iterator[Tuple[federated_data.ClientId, Params]]:
    yield from self._train_each_client(shared_input=params, clients=clients)

  def train_per_client_params(
      self, clients: Iterable[Tuple[federated_data.ClientId,
                                    Iterable[BatchExample], PRNGKey, Params]]
  ) -> Iterator[Tuple[federated_data.ClientId, Params]]:
    yield from self._train_each_client(
        shared_input=None,
        clients=[(client_id, batches, (rng, params))
                 for client_id, batches, rng, params in clients])


class ClientDeltaTrainer(_BaseClientTrainer):
  """Trains each client dataset, starting from either global params, or per client params.

  The result values are delta params (i.e. initial params - final params).
  """

  def __init__(self, grad_fn: Callable[[Params, BatchExample, PRNGKey],
                                       jnp.ndarray],
               client_optimizer: optimizers.Optimizer):
    super().__init__(grad_fn, client_optimizer, return_delta=True)


class ClientParamsTrainer(_BaseClientTrainer):
  """Trains each client dataset, starting from either global params, or per client params.

  The result values are final params.
  """

  def __init__(self, grad_fn: Callable[[Params, BatchExample, PRNGKey],
                                       jnp.ndarray],
               client_optimizer: optimizers.Optimizer):
    super().__init__(grad_fn, client_optimizer, return_delta=False)


def maximization_step(
    evaluator: models.AverageLossEvaluator, cluster_params: List[Params],
    clients: Sequence[Tuple[federated_data.ClientId,
                            client_datasets.ClientDataset, PRNGKey]],
    batch_hparams: client_datasets.PaddedBatchHParams
) -> Dict[federated_data.ClientId, int]:
  """Evaluates and assigns each client to the cluster with the lowest average loss."""
  cluster_losses = _cluster_losses(
      evaluator=evaluator,
      cluster_params=cluster_params,
      clients=clients,
      batch_hparams=batch_hparams)
  return jax.device_get(_cluster_assignment(cluster_losses))


def _cluster_losses(
    evaluator: models.AverageLossEvaluator, cluster_params: List[Params],
    clients: Sequence[Tuple[federated_data.ClientId,
                            client_datasets.ClientDataset, PRNGKey]],
    batch_hparams: client_datasets.PaddedBatchHParams
) -> Dict[federated_data.ClientId, List[jnp.ndarray]]:
  """Evaluates the average loss for each client and cluster."""
  cluster_losses = {client_id: [] for client_id, _, _ in clients}
  client_rngs = [
      jax.random.split(rng, len(cluster_params)) for _, _, rng in clients
  ]
  for i, params in enumerate(cluster_params):
    for client_id, average_loss in evaluator.evaluate_global_params(
        params,
        [(client_id, dataset.padded_batch(batch_hparams), rng[i])
         for (client_id, dataset, _), rng in zip(clients, client_rngs)]):
      cluster_losses[client_id].append(average_loss)
  return cluster_losses


@jax.jit
def _cluster_assignment(cluster_losses):
  """Assigns each client the cluster with the lowest loss."""
  return {
      client_id: jnp.argmin(jnp.stack(losses))
      for client_id, losses in cluster_losses.items()
  }


def expectation_step(trainer: ClientDeltaTrainer, cluster_params: List[Params],
                     client_cluster_ids: Dict[federated_data.ClientId, int],
                     clients: Sequence[Tuple[federated_data.ClientId,
                                             client_datasets.ClientDataset,
                                             PRNGKey]],
                     batch_hparams: client_datasets.ShuffleRepeatBatchHParams):
  """Updates each cluster's params using average delta from clients in this cluster."""
  # Train each client starting from its corresponding cluster params, and
  # calculate weighted average of delta params within each cluster.
  num_examples = {client_id: len(dataset) for client_id, dataset, _ in clients}
  cluster_delta_params_sum = [
      jax.tree_util.tree_map(jnp.zeros_like, params)
      for params in cluster_params
  ]
  cluster_num_examples_sum = [0 for _ in cluster_params]
  for client_id, delta_params in trainer.train_per_client_params([
      (client_id, dataset.shuffle_repeat_batch(batch_hparams), rng,
       cluster_params[client_cluster_ids[client_id]])
      for client_id, dataset, rng in clients
  ]):
    cluster_id = client_cluster_ids[client_id]
    cluster_delta_params_sum[cluster_id] = tree_util.tree_add(
        cluster_delta_params_sum[cluster_id],
        tree_util.tree_weight(delta_params, num_examples[client_id]))
    cluster_num_examples_sum[cluster_id] += num_examples[client_id]
  # Weighted average delta params, or None if no examples were seen for this
  # cluster.
  cluster_delta_params = []
  for delta_params_sum, num_examples_sum in zip(cluster_delta_params_sum,
                                                cluster_num_examples_sum):
    if num_examples_sum > 0:
      cluster_delta_params.append(
          tree_util.tree_inverse_weight(delta_params_sum, num_examples_sum))
    else:
      cluster_delta_params.append(None)
  return cluster_delta_params


@functools.partial(jax.jit, static_argnums=(0, 1))
def random_init(num_clusters: int, init: Callable[[PRNGKey], Params],
                rng: PRNGKey) -> List[Params]:
  """Randomly initializes cluster params."""
  return [init(i) for i in jax.random.split(rng, num_clusters)]


class ModelKMeansInitializer:
  """Initializes cluster params for HypCluster using a k-means++ variant.

  This is a thin wrapper for initializing from a
  :class:`Model <fedjax.core.models.Model>` .
  See :func:`kmeans_init` for a more general version of the initializer, and
  details of the initialization algorithm.
  """

  def __init__(self,
               model: models.Model,
               client_optimizer: optimizers.Optimizer,
               regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
    self._model = model
    self._trainer = ClientParamsTrainer(
        models.model_grad(model, regularizer), client_optimizer)
    self._evaluator = models.AverageLossEvaluator(
        models.model_per_example_loss(model), regularizer)

  def cluster_params(
      self, num_clusters: int, rng: PRNGKey,
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]],
      train_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
      eval_batch_hparams: client_datasets.PaddedBatchHParams) -> List[Params]:
    # Shared initial params.
    init_rng, center_rng = jax.random.split(rng)
    init_params = self._model.init(init_rng)
    return kmeans_init(
        num_clusters=num_clusters,
        init_params=init_params,
        clients=clients,
        trainer=self._trainer,
        train_batch_hparams=train_batch_hparams,
        evaluator=self._evaluator,
        eval_batch_hparams=eval_batch_hparams,
        rng=center_rng)


def kmeans_init(num_clusters: int, init_params: Params,
                clients: Sequence[Tuple[federated_data.ClientId,
                                        client_datasets.ClientDataset,
                                        PRNGKey]], trainer: ClientParamsTrainer,
                train_batch_hparams: client_datasets.ShuffleRepeatBatchHParams,
                evaluator: models.AverageLossEvaluator,
                eval_batch_hparams: client_datasets.PaddedBatchHParams,
                rng: PRNGKey) -> List[Params]:
  """Initializes cluster params for HypCluster using a k-means++ variant.

  See :class:`ModelKMeansInitializer` for a more convenient initializer from a
  :class:`Model <fedjax.core.models.Model>`.

  Given a set of input clients, we train parameters for each client. The
  initial cluster parameters are chosen out of this set of client parameters.
  At a high level, the cluster parameters are selected by choosing the client
  parameters that are the "furthest away" from each other (greatest difference
  in loss).

  Args:
    num_clusters: Desired number of cluster centers.
    init_params: Initial params for training each client.
    clients: Clients to train for generating candidate cluster center params.
    trainer: Client trainer for carry out client training.
    train_batch_hparams: Batching hyperparameters for client training.
    evaluator: Average loss evaluator for evaluator clients on chosen cluster
      center params.
    eval_batch_hparams: Batching hyperparameters for client evaluation.
    rng: RNG used for choosing the initial cluster center.

  Returns:
    Cluster center params.

  Raises:
    ValueError if some input arguments are invalid.
  """
  if num_clusters <= 0:
    raise ValueError(f'num_clusters should be > 0, got {num_clusters}')
  # Choose a client to be the initial center.
  initial_center_index = jax.random.choice(rng, len(clients))
  del rng  # Free up the name "rng" for subsequent loops.

  # Split client RNGs. For each client, we need 1 RNG for training,
  # (num_clusters - 1) for average loss evaluation.
  client_rngs = [jax.random.split(rng, num_clusters) for _, _, rng in clients]

  # Train each client from init_params.
  train_clients = [
      (client_id, dataset.shuffle_repeat_batch(train_batch_hparams), rng[0])
      for (client_id, dataset, _), rng in zip(clients, client_rngs)
  ]
  client_params = dict(trainer.train_global_params(init_params, train_clients))

  # TODO(wuke): Should clients already chosen be eliminated from subsequent
  # evaluations to prevent the same center from being added multiple times?

  # Choose cluster centers, starting with a randomly chosen initial center.
  cluster_centers = [client_params[sorted(client_params)[initial_center_index]]]
  # Repeatedly add the client with the worse best loss as the next center.
  best_losses = {client_id: jnp.inf for client_id in client_params}
  for i in range(1, num_clusters):
    # Evaluate all clients with the newly added center.
    eval_clients = [
        (client_id, dataset.padded_batch(eval_batch_hparams), rng[i])
        for (client_id, dataset, _), rng in zip(clients, client_rngs)
    ]
    new_losses = dict(
        evaluator.evaluate_global_params(cluster_centers[-1], eval_clients))
    # Update best loss for each client.
    best_losses = jax.tree_util.tree_multimap(min, best_losses,
                                              jax.device_get(new_losses))
    # Add a new center.
    worst_client = max(best_losses, key=lambda k: best_losses[k])
    cluster_centers.append(client_params[worst_client])

  return cluster_centers


class HypClusterEvaluator:
  """Evaluates cluster params on a Model."""

  def __init__(self,
               model: models.Model,
               regularizer: Optional[Callable[[Params], jnp.ndarray]] = None):
    """Initializes some reusable components.

    Because we need to make 2 passes over each client during evaluation, the
    number of clients that can be evaluated at once is limited. Therefore
    multiple calls to :meth:`~HypClusterEvaluator.evaluate_clients` are needed
    to evaluate a large federated dataset. We factor out some reusable
    components so that the same computation can be jit compiled.

    Args:
      model: Model being evaluated.
      regularizer: Optional regularizer.
    """
    self._maximization_step_evaluator = models.AverageLossEvaluator(
        models.model_per_example_loss(model), regularizer)
    self._model_evaluator = models.ModelEvaluator(model)

  def evaluate_clients(
      self, cluster_params: List[Params],
      clients: Sequence[Tuple[federated_data.ClientId,
                              client_datasets.ClientDataset, PRNGKey]],
      batch_hparams: client_datasets.PaddedBatchHParams
  ) -> Iterator[Tuple[federated_data.ClientId, Dict[str, jnp.ndarray]]]:
    """Evaluates each client on the cluster with best average loss."""
    maximization_clients = clients
    cluster_client_ids = maximization_step(
        evaluator=self._maximization_step_evaluator,
        cluster_params=cluster_params,
        clients=maximization_clients,
        batch_hparams=batch_hparams)
    eval_clients = [(client_id, dataset.padded_batch(batch_hparams),
                     cluster_params[cluster_client_ids[client_id]])
                    for client_id, dataset, _ in clients]
    yield from self._model_evaluator.evaluate_per_client_params(eval_clients)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.datasets.toy_regression."""

from fedjax.legacy.datasets import toy_regression
import tensorflow as tf


class ToyRegressionDataTest(tf.test.TestCase):

  def test_load_data(self):
    num_clients = 10
    train_data, test_data = toy_regression.load_data(
        num_clients=num_clients, num_domains=2, num_points=100, seed=10)
    client_id = train_data.client_ids[3]

    train_client_data = list(train_data.create_tf_dataset_for_client(client_id))
    test_client_data = list(test_data.create_tf_dataset_for_client(client_id))

    self.assertLen(train_data.client_ids, num_clients)
    self.assertEqual(train_data.client_ids, test_data.client_ids)
    self.assertNotAllEqual(train_client_data[0]['y'], test_client_data[0]['y'])


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated EMNIST."""

from typing import Optional, Tuple

from fedjax.core import client_datasets
from fedjax.core import federated_data
from fedjax.core import sqlite_federated_data
from fedjax.datasets import downloads
import numpy as np

SPLITS = ('train', 'test')


def cite():
  return """@inproceedings{cohen2017emnist,
  title={EMNIST: Extending MNIST to handwritten letters},
  author={Cohen, Gregory and Afshar, Saeed and Tapson, Jonathan and
Van Schaik, Andre},
  booktitle={2017 International Joint Conference on Neural Networks (IJCNN)},
  pages={2921--2926},
  year={2017},
  organization={IEEE}
}"""


def load_split(split: str,
               only_digits: bool = False,
               mode: str = 'sqlite',
               cache_dir: Optional[str] = None) -> federated_data.FederatedData:
  """Loads an unprocessed federated emnist split.

  Features:
    pixels: [N, 28, 28] float32 image pixels.
    label: [N] int32 classification label.

  Args:
    split: Name of the split. One of SPLITS.
    only_digits: Whether to only load the digits data.
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    FederatedData.
  """
  if split not in SPLITS:
    raise ValueError(f'Invalid split={split!r}')
  if cache_dir is not None and mode != 'sqlite':
    raise ValueError('Caching locally is only supported in "sqlite" mode')
  if only_digits:
    name = 'digitsonly_' + split
  else:
    name = split
  if mode == 'sqlite':
    path = downloads.maybe_download(
        f'https://storage.googleapis.com/gresearch/fedjax/emnist/federated_emnist_{name}.sqlite',
        cache_dir)
    return sqlite_federated_data.SQLiteFederatedData.new(path)
  else:
    raise ValueError(f'Unsupported mode={mode!r}')


def domain_id(client_id: federated_data.ClientId) -> int:
  """Returns domain id for client id."""
  # client ids are of the following format:
  # - sqlite: "[16-byte hex hash]:f[4-digit integer]_[2-digit integer]"
  #
  # These domain ids are based on NIST data source. For more details, see
  # https://s3.amazonaws.com/nist-srd/SD19/sd19_users_guide_edition_2.pdf.
  if len(client_id) == 25:
    cid = int(client_id[18:22])
  elif len(client_id) == 8:
    cid = int(client_id[1:5])
  else:
    raise ValueError(f'Invalid client_id: {client_id!r}')
  if 2100 <= cid and cid <= 2599:
    return 0  # HIGH_SCHOOL.
  return 1  # CENSUS_FIELD.


def preprocess_client(
    client_id: federated_data.ClientId,
    examples: client_datasets.Examples) -> client_datasets.Examples:
  return {
      **examples, 'domain_id':
          np.full_like(examples['label'], domain_id(client_id))
  }


def preprocess_batch(
    examples: client_datasets.Examples) -> client_datasets.Examples:
  return {
      'x': 1 - examples['pixels'][..., np.newaxis],
      'y': examples['label'],
      'domain_id': examples['domain_id']
  }


def preprocess_split(
    fd: federated_data.FederatedData) -> federated_data.FederatedData:
  return (fd.preprocess_client(preprocess_client).preprocess_batch(
      preprocess_batch))


def load_data(
    only_digits: bool = False,
    mode: str = 'sqlite',
    cache_dir: Optional[str] = None
) -> Tuple[federated_data.FederatedData, federated_data.FederatedData]:
  """Loads processed EMNIST train and test splits.

  Features:
    x: [N, 28, 28, 1] float32 flipped image pixels.
    y: [N] int32 classification label.
    domain_id: [N] int32 domain id (see domain_id()).

  Args:
    only_digits: Whether to only load the digits data.
    mode: 'sqlite'.
    cache_dir: Directory to cache files in 'sqlite' mode.

  Returns:
    Train and test splits as FederatedData.
  """
  train = load_split(
      'train', only_digits=only_digits, mode=mode, cache_dir=cache_dir)
  test = load_split(
      'test', only_digits=only_digits, mode=mode, cache_dir=cache_dir)
  return preprocess_split(train), preprocess_split(test)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Federated experiment manager."""

import abc
import os.path
import time
from typing import Any, Mapping, NamedTuple, Optional, Sequence, Tuple

from absl import logging
from fedjax.core import client_datasets
from fedjax.core import client_samplers
from fedjax.core import federated_algorithm
from fedjax.core import federated_data
from fedjax.core import models
from fedjax.training import checkpoint
from fedjax.training import logging as fedjax_logging
import jax.numpy as jnp
import tensorflow as tf


def set_tf_cpu_only():
  """Restricts TensorFlow device visibility to only CPU.

  TensorFlow is only used for data loading, so we prevent it from allocating
  GPU/TPU memory.
  """
  tf.config.experimental.set_visible_devices([], 'GPU')
  tf.config.experimental.set_visible_devices([], 'TPU')


class EvaluationFn(metaclass=abc.ABCMeta):
  """Evaluation function that are only fed state at every call.

  Typically used for full evaluation or evaluation on sampled clients from a
  test set.
  """

  @abc.abstractmethod
  def __call__(self, state: federated_algorithm.ServerState,
               round_num: int) -> Mapping[str, jnp.ndarray]:
    """Runs final evaluation."""


class ModelSampleClientsEvaluationFn(EvaluationFn):
  """Evaluation on sampled clients using the centralized model.

  The state to be evaluated must contain a params field.
  """

  def __init__(self, client_sampler: client_samplers.ClientSampler,
               model: models.Model,
               batch_hparams: client_datasets.PaddedBatchHParams):
    self._client_sampler = client_sampler
    self._model = model
    self._batch_hparams = batch_hparams

  def __call__(self, state: federated_algorithm.ServerState,
               round_num: int) -> Mapping[str, jnp.ndarray]:
    params = state.params
    self._client_sampler.set_round_num(round_num)
    clients = self._client_sampler.sample()
    batches = client_datasets.padded_batch_client_datasets(
        [i for _, i, _ in clients], self._batch_hparams)
    return models.evaluate_model(self._model, params, batches)


class ModelFullEvaluationFn(EvaluationFn):
  """Evaluation on an entire federated dataset using the centralized model."""

  def __init__(self, fd: federated_data.FederatedData, model: models.Model,
               batch_hparams: client_datasets.PaddedBatchHParams):
    self._fd = fd
    self._model = model
    self._batch_hparams = batch_hparams

  def __call__(self, state: federated_algorithm.ServerState,
               round_num: int) -> Mapping[str, jnp.ndarray]:
    del round_num
    params = state.params
    batches = federated_data.padded_batch_federated_data(
        self._fd, self._batch_hparams)
    return models.evaluate_model(self._model, params, batches)


class TrainClientsEvaluationFn(metaclass=abc.ABCMeta):
  """Evaluation function that are fed training clients at every call.

  Typically used for evaluation on the training clients used in a step.
  """

  @abc.abstractmethod
  def __call__(
      self, state: federated_algorithm.ServerState, round_num: int,
      train_clients: Sequence[Tuple[federated_data.ClientId,
                                    client_datasets.ClientDataset, Any]]
  ) -> Mapping[str, jnp.ndarray]:
    """Runs evaluation."""


class ModelTrainClientsEvaluationFn(TrainClientsEvaluationFn):
  """Evaluation on training clients using the centralized model.

  The state to be evaluated must contain a params field.
  """

  def __init__(self, model: models.Model,
               batch_hparams: client_datasets.PaddedBatchHParams):
    self._model = model
    self._batch_hparams = batch_hparams

  def __call__(
      self, state: federated_algorithm.ServerState, round_num: int,
      train_clients: Sequence[Tuple[federated_data.ClientId,
                                    client_datasets.ClientDataset, Any]]
  ) -> Mapping[str, jnp.ndarray]:
    del round_num
    params = state.params
    batches = client_datasets.padded_batch_client_datasets(
        [client_dataset for _, client_dataset, _ in train_clients],
        self._batch_hparams)
    return models.evaluate_model(self._model, params, batches)


class FederatedExperimentConfig(NamedTuple):
  """Common configurations of a federated experiment.

  Attribues:
    root_dir: Root directory for experiment outputs (e.g. metrics).
    num_rounds: Number of federated training rounds.
    checkpoint_frequency: Checkpoint frequency in rounds. If <= 0, no
    checkpointing is done.
    num_checkpoints_to_keep: Maximum number of checkpoints to keep.
    eval_frequency: Evaluation frequency in rounds. If <= 0, no evaluation is
    done.
  """
  root_dir: str
  num_rounds: int
  checkpoint_frequency: int = 0
  num_checkpoints_to_keep: int = 1
  eval_frequency: int = 0


def run_federated_experiment(
    algorithm: federated_algorithm.FederatedAlgorithm,
    init_state: federated_algorithm.ServerState,
    client_sampler: client_samplers.ClientSampler,
    config: FederatedExperimentConfig,
    periodic_eval_fn_map: Optional[Mapping[str, Any]] = None,
    final_eval_fn_map: Optional[Mapping[str, EvaluationFn]] = None
) -> federated_algorithm.ServerState:
  """Runs the training loop of a federated algorithm experiment.

  Args:
    algorithm: Federated algorithm to use.
    init_state: Initial server state.
    client_sampler: Sampler for training clients.
    config: FederatedExperimentConfig configurations.
    periodic_eval_fn_map: Mapping of name to evaluation functions that are run
      repeatedly over multiple federated training rounds. The frequency is
      defined in `_FederatedExperimentConfig.eval_frequency`.
    final_eval_fn_map: Mapping of name to evaluation functions that are run at
      the very end of federated training. Typically, full test evaluation
      functions will be set here.

  Returns:
    Final state of the input federated algortihm after training.
  """
  if config.root_dir:
    tf.io.gfile.makedirs(config.root_dir)

  if periodic_eval_fn_map is None:
    periodic_eval_fn_map = {}
  if final_eval_fn_map is None:
    final_eval_fn_map = {}

  logger = fedjax_logging.Logger(config.root_dir)

  latest = checkpoint.load_latest_checkpoint(config.root_dir)
  if latest:
    state, last_round_num = latest
    start_round_num = last_round_num + 1
  else:
    state = init_state
    start_round_num = 1
  client_sampler.set_round_num(start_round_num)

  start = time.time()
  for round_num in range(start_round_num, config.num_rounds + 1):
    # Get a random state and randomly sample clients.
    clients = client_sampler.sample()
    client_ids = [i[0] for i in clients]
    logging.info('round_num %d: client_ids = %s', round_num, client_ids)

    # Run one round of the algorithm, where bulk of the work happens.
    state, _ = algorithm.apply(state, clients)

    # Save checkpoint.
    should_save_checkpoint = config.checkpoint_frequency and (
        round_num == start_round_num or
        round_num % config.checkpoint_frequency == 0)
    if should_save_checkpoint:
      checkpoint.save_checkpoint(config.root_dir, state, round_num,
                                 config.num_checkpoints_to_keep)

    # Run evaluation.
    should_run_eval = config.eval_frequency and (
        round_num == start_round_num or round_num % config.eval_frequency == 0)
    if should_run_eval:
      start_periodic_eval = time.time()
      for eval_name, eval_fn in periodic_eval_fn_map.items():
        if isinstance(eval_fn, EvaluationFn):
          metrics = eval_fn(state, round_num)
        elif isinstance(eval_fn, TrainClientsEvaluationFn):
          metrics = eval_fn(state, round_num, clients)
        else:
          raise ValueError(f'Invalid eval_fn type {type(eval_fn)}')
        if metrics:
          for metric_name, metric_value in metrics.items():
            logger.log(eval_name, metric_name, metric_value, round_num)
      logger.log('.', 'periodic_eval_duration_sec',
                 time.time() - start_periodic_eval, round_num)

    # Log the time it takes per round. Rough approximation since we're not
    # using DeviceArray.block_until_ready()
    logger.log('.', 'mean_round_duration_sec',
               (time.time() - start) / (round_num + 1 - start_round_num),
               round_num)

  # Block until previous work has finished.
  jnp.zeros([]).block_until_ready()

  # Logging overall time it took.
  num_rounds = config.num_rounds - start_round_num + 1
  mean_round_duration = ((time.time() - start) /
                         num_rounds if num_rounds > 0 else 0)

  # Final evaluation.
  final_eval_start = time.time()
  for eval_name, eval_fn in final_eval_fn_map.items():
    metrics = eval_fn(state, round_num)
    if metrics:
      metrics_path = os.path.join(config.root_dir, f'{eval_name}.tsv')
      with tf.io.gfile.GFile(metrics_path, 'w') as f:
        f.write('\t'.join(metrics.keys()) + '\n')
        f.write('\t'.join([str(v) for v in metrics.values()]))
  # DeviceArray.block_until_ready() isn't needed here since we write to file.
  final_eval_duration = time.time() - final_eval_start
  logging.info('mean_round_duration = %f sec.', mean_round_duration)
  logging.info('final_eval_duration = %f sec.', final_eval_duration)
  return state



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.legacy.core.test_util."""

from fedjax.legacy.core import test_util
import haiku as hk
import tensorflow as tf


class TestUtilsTest(tf.test.TestCase):

  def test_create_toy_data(self):
    data = test_util.create_toy_data(
        num_clients=10,
        num_clusters=2,
        num_classes=4,
        num_examples=5,
        seed=10)
    client_id = data.client_ids[3]
    client_data = list(data.create_tf_dataset_for_client(client_id))
    self.assertLen(data.client_ids, 10)
    self.assertLen(client_data, 5)
    self.assertAllEqual(client_data[0]['x'].shape, [1])
    self.assertAllEqual(client_data[0]['y'].shape, [])

  def test_create_toy_model(self):
    model = test_util.create_toy_model(
        num_classes=10, num_hidden_layers=2, feature_dim=5)
    params = model.init_params(next(hk.PRNGSequence(0)))
    self.assertEqual(list(params.keys()), ['linear', 'linear_1', 'linear_2'])
    self.assertEqual(hk.data_structures.tree_size(params), 750)

  def test_create_toy_example(self):
    data, model = test_util.create_toy_example(
        num_clients=10,
        num_clusters=2,
        num_classes=4,
        num_examples=5,
        seed=10)
    batch = next(
        (data.create_tf_dataset_for_client(data.client_ids[0])
         .batch(3)
         .as_numpy_iterator()))
    params = model.init_params(next(hk.PRNGSequence(0)))
    self.assertTupleEqual(model.apply_fn(params, None, batch).shape, (3, 4))

  def test_create_mock_state(self):
    state = test_util.create_mock_state(seed=0)
    self.assertEqual(list(state.params.keys()), ['linear'])
    self.assertEqual(hk.data_structures.tree_size(state.params), 124)

  def test_mock_federated_algorithm(self):
    num_clients = 10
    num_examples = 5
    num_rounds = 3
    algorithm = test_util.MockFederatedAlgorithm(
        num_clients=num_clients, num_examples=num_examples)

    state = algorithm.init_state()
    for _ in range(num_rounds):
      state = algorithm.run_round(state, algorithm.federated_data.client_ids)

    self.assertEqual(state.count, num_rounds * num_clients * num_examples)


if __name__ == '__main__':
  tf.test.main()
